import numpy as np
import sys, os, time
import torch 
import torch.nn as nn
import torch.backends.cudnn as cudnn
sys.path.append('../')
import data as data
import architecture as architecture

##################################### INPUT ##########################################
root_in      = '/mnt/ceph/users/camels'
root_out     = '/mnt/ceph/users/camels/Results/neural_nets/params_2_SFRH'
sim          = 'IllustrisTNG'
seed         = 1
realizations = 1000
batch_size   = 25
bins_SFRH    = 100

# architecture parameters
h1 = 500
h2 = 100
h3 = 100
h4 = 100
wd = 1e-7
dr = 0.2

# training parameters
lr     = 2e-4
epochs = 100000

fout   = '%s/losses.txt'%root_out
fmodel = '%s/model.pt'%root_out
######################################################################################

# use GPUs if available
if torch.cuda.is_available():
    print("CUDA Available")
    device = torch.device('cuda')
else: 
    print("CUDA Not Available")
    device = torch.device('cpu')
cudnn.benchmark = True   

# define loss function
criterion = nn.MSELoss() 

# get train and validation sets
print('preparing dataset loaders')
train_loader, valid_loader = data.create_train_valid_sets(seed, realizations, root_in, 
                                                bins_SFRH, sim, batch_size, root_out)
train_batches = len(train_loader)
valid_batches = len(valid_loader)

# get test set
test_loader = data.create_test_set(seed, realizations, root_in, bins_SFRH, sim,
                                   batch_size, root_out)
test_batches = len(test_loader)

# define architecture
model = architecture.model_1hl(6, h1, bins_SFRH, dr)
#model = architecture.model_3hl(6, h1, h2, h3, bins_SFRH, dr)
model.to(device=device)
network_total_params = sum(p.numel() for p in model.parameters())
print('total number of parameters in the model = %d'%network_total_params)

# define optimizer and scheduler
optimizer = torch.optim.Adam(model.parameters(), lr=lr, betas=(0.5, 0.999), 
                             weight_decay=wd)	
scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', 
                                factor=0.5, patience=25, verbose=True, eps=1e-8)

# load best-model, if it exists
if os.path.exists(fmodel):  
    print('Loading model...')
    model.load_state_dict(torch.load(fmodel))
    model.to(device=device)

# get validation loss
print('Computing initial validation loss')
model.eval()
min_valid_loss = 0.0
for params_val, SFRH_val in valid_loader:
    with torch.no_grad():
        params_val = params_val.to(device=device)
        SFRH_val   = SFRH_val.to(device=device)
        SFRH_pred = model(params_val)
        min_valid_loss += criterion(SFRH_pred, SFRH_val).item()
min_valid_loss /= valid_batches
print('Initial valid loss = %.3e'%min_valid_loss)

# see if results for this model are available
if os.path.exists(fout):  
    offset, dumb, dumb = np.loadtxt(fout, unpack=True)
    offset = int(offset[-1]+1)
else:  offset = 0

# do a loop over all epochs
start = time.time()
for epoch in range(offset,offset+epochs):
        
    # do training
    train_loss, count = 0.0, 0
    model.train()
    for params_train, SFRH_train in train_loader:
        params_train = params_train.to(device)
        SFRH_train   = SFRH_train.to(device)
        SFRH_pred  = model(params_train)

        loss = criterion(SFRH_pred, SFRH_train)
        train_loss += loss.item()
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        count += 1
        if count>train_batches:  break
    train_loss /= train_batches

    # do validation
    valid_loss = 0.0
    model.eval()
    for params_val, SFRH_val in valid_loader:
        with torch.no_grad():
            params_val = params_val.to(device)
            SFRH_val   = SFRH_val.to(device)
            SFRH_pred = model(params_val)
            valid_loss += criterion(SFRH_pred, SFRH_val).item()
    valid_loss /= valid_batches

    # do testing
    test_loss = 0.0
    model.eval()
    for params_test, SFRH_test in test_loader:
        with torch.no_grad():
            params_test = params_test.to(device)
            SFRH_test   = SFRH_test.to(device)
            SFRH_pred = model(params_test)
            test_loss += criterion(SFRH_pred, SFRH_test).item()
    test_loss /= test_batches

    #scheduler.step(valid_loss)

    # save model if it is better
    if valid_loss<min_valid_loss:
        print('saving model...')
        torch.save(model.state_dict(), fmodel)
        min_valid_loss = valid_loss
        print('%03d %.3e %.3e %.3e'%(epoch, train_loss, valid_loss, test_loss))
    
    # save losses to file
    f = open(fout, 'a')
    f.write('%d %.5e %.5e %.5e\n'%(epoch, train_loss, valid_loss, test_loss))
    f.close()
    
stop = time.time()
print('Time take (m):', "{:.4f}".format((stop-start)/60.0))



import numpy as np
import sys,os
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.optim as optim
import torchvision.utils as vutils
import data
import architecture

################################### INPUT ############################################
root_out   = '/mnt/ceph/users/camels/Results/autoencoder'
seed       = 5 #to split images between training, validation and testing sets
grid       = 64
batch_size = 128 
BN_dim     = 500
hidden     = 32
examples   = 16
sim        = 'IllustrisTNG_fiducial' #'IllustrisTNG', 'SIMBA', 'IllustrisTNG_fiducial'
mode       = 'all'  #'train', 'valid', 'test', 'all'

#f_model    = 'model_a_32_wd=1e-9_noise_500dim.pt'
#minimum, maximum = 3.182, 7.948

f_model    = 'model_fiducial_a_32_wd=1e-10_500dim.pt'
minimum, maximum = 3.277, 7.507 

fout = '%s/results/Errors_train_on_fiducialTNG_test_on_allfiducialTNG.txt'%root_out
######################################################################################

# get the name of the file containing the images and the data
if sim=='IllustrisTNG':  
    f_images    = '/mnt/ceph/users/camels/Results/GAN/Images_T.npy'
    test_loader = data.create_dataset(mode, seed, f_images, grid, batch_size, 
                                      minimum, maximum, verbose=True)
elif sim=='SIMBA':
    f_images    = '/mnt/ceph/users/camels/Results/GAN/Images_T_SIMBA.npy'
    test_loader = data.create_dataset(mode, seed, f_images, grid, batch_size, 
                                      minimum, maximum, verbose=True)
elif sim=='IllustrisTNG_fiducial':
    f_images    = '/mnt/ceph/users/camels/Results/autoencoder/Images_T_fiducial_IllustrisTNG.npy'
    test_loader = data.create_dataset_fiducial(mode, seed, f_images, grid, batch_size, 
                                               minimum, maximum, verbose=True)
else:
    raise Exception('Incorrect sim!')

# use GPUs if available
GPU = torch.cuda.is_available()
device = torch.device("cuda" if GPU else "cpu")
print('GPU:',GPU)
print('Training on',device)
cudnn.benchmark = True      #May train faster but cost more memory

# define loss function
criterion = nn.MSELoss()

# define the model and load best model
model = architecture.autoencoder_64a(BN_dim, hidden).to(device)
model.load_state_dict(torch.load('%s/models/%s'%(root_out, f_model)))

# define variables needed to compute the error histogram
bins_histo  = np.logspace(-6,-2, 301)
mean_histo  = 10**(0.5*(np.log10(bins_histo[1:]) + np.log10(bins_histo[:-1])))
dhisto      = bins_histo[1:] - bins_histo[:-1]
histo_error = np.zeros(bins_histo.shape[0]-1, dtype=np.float64)

# compute test loss
test_loss, maps, batches = 0.0, 0, 0
min_error, mean_error, max_error = 1e9, 0.0, 0.0
model.eval()
for test_maps, numbers in test_loader:
    with torch.no_grad():

        # get input, output maps and loss
        test_maps = test_maps.to(device)
        recon_maps = model(test_maps)
        loss = criterion(recon_maps, test_maps)

        # get cumulative loss and number of maps
        test_loss += loss.cpu().item()
        batches += 1
        maps += test_maps.shape[0]

        # save some images
        #if test_maps.shape[0]>examples:
        #    vutils.save_image(torch.cat((test_maps[:examples],recon_maps[:examples])), 
        #                      '%s/images/images_%s_%s.png'%(root_out,sim,mode),\
        #                      normalize=True, nrow=examples, range=(-1.0, 1.0))

        # compute reconstruction error for each map
        error = ((test_maps - recon_maps)**2).cpu().numpy()
        error = np.mean(error, axis=(1,2,3))
        mean_error += np.sum(error)
        histo_error += np.histogram(error, bins_histo)[0]

        # identify the larger reconstruction error and make image
        index = np.where(error==np.max(error))[0]
        if error[index]>max_error:
            max_error = error[index]
            print('max error = %.3e : %d'%(max_error, numbers[index]))
            #vutils.save_image(torch.cat([test_maps[index],recon_maps[index]]), 
            #                 '%s/images/images_%s_%s_outlier.png'%(root_out,sim,mode),\
            #                  normalize=True, nrow=1, range=(-1.0, 1.0))

        # identify the smaller reconstruction error
        index = np.where(error==np.min(error))[0]
        if error[index]<min_error:
            min_error = error[index]
            print('min error = %.3e : %d'%(min_error, numbers[index]))

# final verbose
print('\nTest loss  = %.3e'%(test_loss/batches))
print('Min = %.3e  :  Max = %.3e'%(min_error, max_error))
print('Mean error = %.3e'%(mean_error/maps))
print('Number of maps = %d'%maps)

# histogram
histo_error = histo_error/maps#/dhisto
print('Integral histo = %.3e'%(np.sum(histo_error*dhisto)))
np.savetxt(fout, np.transpose([mean_histo, histo_error]))



import numpy as np
import sys,os
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.optim as optim
import torchvision.utils as vutils
import data
import architecture
import matplotlib.pyplot as plt
import utils


################################### INPUT ############################################
# images parameters
f_images = '/mnt/ceph/users/camels/Results/autoencoder/Images_T_fiducial_IllustrisTNG.npy'
seed     = 5 #to split images between training, validation and testing sets
grid     = 64
examples = 16 # when saving images for different epochs

# architecture parameters
lr         = 5e-5
hidden     = 64
wd         = 1e-10
epochs     = 100000
batch_size = 128
BN_dim     = 500

# output files
root_out = '/mnt/ceph/users/camels/Results/autoencoder'
f_loss   = '%s/losses/loss_fiducial_model_a_64_wd=1e-10_500dim.txt'%root_out
f_model  = '%s/models/model_fiducial_a_64_wd=1e-10_500dim.pt'%root_out
######################################################################################

# use GPUs if available
GPU = torch.cuda.is_available()
device = torch.device("cuda" if GPU else "cpu")
print('GPU:',GPU)
print('Training on',device)
cudnn.benchmark = True      #May train faster but cost more memory

# define loss function
criterion = nn.MSELoss()
#criterion = nn.L1Loss()

# get the data
train_loader = data.create_dataset_fiducial('train', seed, f_images, grid, batch_size, 
                                            verbose=True)
valid_loader = data.create_dataset_fiducial('valid', seed, f_images, grid, batch_size, 
                                            verbose=True)

# define the model
model = architecture.autoencoder_64a(BN_dim, hidden).to(device)
#model = architecture.autoencoder_64h(BN_dim, hidden).to(device)
#model = architecture.autoencoder_64d(BN_dim, hidden).to(device)
#model = architecture.autoencoder_64e(BN_dim, hidden).to(device)
#model = architecture.autoencoder_64f(BN_dim, hidden).to(device)
#model = architecture.autoencoder_64g(BN_dim, hidden).to(device)
#model = architecture.autoencoder_64c(BN_dim, hidden).to(device)
#model = architecture.autoencoder_64b(BN_dim, hidden).to(device)

# load best-models, if they exists
if os.path.exists(f_model):  model.load_state_dict(torch.load(f_model))
else:                        model.apply(architecture.weights_init)

# define the optimizer
optimizer = optim.Adam(model.parameters(), lr=lr, betas=(0.5, 0.999), weight_decay=wd)

# get the tensor for the images
for valid_maps,idxs in valid_loader:
    test_maps = (valid_maps[:examples]).to(device)
    break

offset = 0
if os.path.exists(f_loss):
    dumb = np.loadtxt(f_loss)
    offset = int(dumb[:,0][-1])

# do a loop over all the epochs
min_loss = 1e8
for epoch in range(offset,offset+epochs):
    
    # training
    train_loss, num_maps = 0.0, 0
    model.train()
    for train_maps,idxs in train_loader:
        train_maps = train_maps.to(device)
        recon_maps = model(train_maps)
        #input_maps = train_maps + torch.randn_like(train_maps, device=device)*0.1
        #recon_maps = model(input_maps)

        #train_maps_fft = torch.rfft(train_maps, 2)
        #recon_maps_fft = torch.rfft(recon_maps, 2)
        #loss = ((train_maps_fft-recon_maps_fft)**2).mean()

        loss = criterion(recon_maps, train_maps)
        train_loss += (loss.cpu().item())*train_maps.shape[0]
        num_maps   += train_maps.shape[0]
        optimizer.zero_grad()
        loss.backward()
        #if epoch%25==0:
        #    utils.save_gradients('%s/gradients/gradients_g_%d.txt'%(root_out,epoch), model)
        optimizer.step()
    train_loss = train_loss/num_maps

    # validation
    valid_loss, num_maps = 0.0, 0
    model.eval()
    for valid_maps,idxs in valid_loader:
        with torch.no_grad():
            valid_maps = valid_maps.to(device)
            recon_maps = model(valid_maps)
            loss = criterion(recon_maps, valid_maps)
            valid_loss += (loss.cpu().item())*valid_maps.shape[0]
            num_maps   += valid_maps.shape[0]
    valid_loss = valid_loss/num_maps

    # verbose
    if valid_loss<min_loss:
        min_loss = valid_loss
        torch.save(model.state_dict(), f_model)
        print('Epoch %d: %.3e %.3e (saving)'%(epoch, train_loss, valid_loss))
    else:
        print('Epoch %d: %.3e %.3e'%(epoch, train_loss, valid_loss))

    # save losses
    f = open(f_loss, 'a')  
    f.write('%d %.3e %.3e\n'%(epoch, train_loss, valid_loss))
    f.close()

    """
    # plot some images & save models every 10 epochs
    if epoch%25==0:
        model.eval()
        with torch.no_grad():
            recon_maps = model(test_maps)
            vutils.save_image(torch.cat([test_maps,recon_maps]), 
                              '%s/images/images_g_%d.png'%(root_out,epoch),\
                              normalize=True, nrow=examples, range=(-1.0, 1.0))
    """



# This script creates images from the extreme IllustrisTNG simulation
import numpy as np
import sys,os,h5py
import MAS_library as MASL
sys.path.append('../analysis')
import camel_library as CL
import torch
import torchvision.utils as vutils

#################################### INPUT ##########################################
root_in      = '/mnt/ceph/users/camels'
root_out     = '/mnt/ceph/users/camels/Results/autoencoder'
sim          = 'SIMBA' #'IllustrisTNG'
snapnum      = 33
realizations = 1000

grid         = 250
splits       = 5

# parameters of the density field routine
periodic = True
x_min, y_min = 0.0, 0.0

# name of snapshot and output file
#snapshot = '%s/Sims/IllustrisTNG/extremestellar_0/snap_%03d.hdf5'%(root_in,snapnum)
#fout     = '%s/Images_T_extremestellar_0.npy'%root_out
#snapshot = '%s/Sims/IllustrisTNG/noFB_0/snap_%03d.hdf5'%(root_in,snapnum)
#fout     = '%s/Images_T_noFB_0.npy'%root_out
snapshot = '%s/Sims/IllustrisTNG/fiducial_0/snap_%03d.hdf5'%(root_in,snapnum)
fout     = '%s/Images_T_fiducial_0.npy'%root_out
#####################################################################################

# define the matrix hosting all the maps
maps_total = np.zeros((splits*3, grid, grid), dtype=np.float32)

# read the gas positions, radii and temperature of gas particles
if not(os.path.exists(snapshot)):  raise Exception('snapshot does not exists')
f = h5py.File(snapshot, 'r')
BoxSize = f['Header'].attrs[u'BoxSize']/1e3 #Mpc/h
pos     = f['PartType0/Coordinates'][:]/1e3 #Mpc/h
mass    = f['PartType0/Masses'][:]*1e10     #Msun/h
radius  = f['PartType0/SubfindHsml'][:]/1e3 #Mpc/h
#radius  = f['PartType0/SmoothingLength'][:]/1e3 #Mpc/h
#radius  = f['PartType0/AGS-Softening'][:]/1e3 #Mpc/h
f.close()
T = CL.temperature(snapshot)

# do a loop over the three different axes
count = 0
for axis in [0,1,2]:

    axis_x, axis_y = (axis+1)%3, (axis+2)%3 

    # do a loop over the different slices of each axis
    for j in range(splits):

        # get the index of the map
        count = splits*axis + j

        # find the range in the slice
        minimum, maximum = j*BoxSize/splits, (j+1)*BoxSize/splits

        # select the particles in the considered slice
        indexes = np.where((pos[:,axis]>=minimum) & (pos[:,axis]<maximum))[0] 
        pos_slice    = pos[indexes].astype(np.float32)
        radius_slice = radius[indexes]
        mass_slice   = mass[indexes]
        T_slice      = T[indexes]

        # project particle mass*temperatures into a 2D maps
        TM = np.zeros((grid,grid), dtype=np.float64)
        MASL.voronoi_RT_2D(TM, pos_slice, T_slice*mass_slice, radius_slice, 
                x_min, y_min, axis_x, axis_y, BoxSize, periodic, verbose=False)

        # project particle mass into a 2D maps
        M = np.zeros((grid,grid), dtype=np.float64)
        MASL.voronoi_RT_2D(M, pos_slice, mass_slice, radius_slice, x_min, y_min,
                           axis_x, axis_y, BoxSize, periodic, verbose=False)

        maps_total[count] = TM/M
print('%d maps'%(count+1))
print(maps_total.shape)
print('%.3e < T [K] < %.3e'%(np.min(maps_total), np.max(maps_total)))
np.save(fout, maps_total)

# save images
maps_total = np.log10(maps_total)
maps_total = torch.tensor(maps_total, dtype=torch.float32)
maps_total = maps_total.unsqueeze(1)
print(maps_total.shape)
vutils.save_image(maps_total, 'images_fiducial_0.png',
                  normalize=True, nrow=5)



import torch 
from torch.utils.data.dataset import Dataset
from torch.utils.data import DataLoader
import numpy as np
import sys, os, time


# This routine reads all SFRH
def read_all_SFRH(root_in, sim, realizations, bins, root_out):

    # check if file already exists
    fout = '%s/SFRH_%s.npy'%(root_out,sim)
    if os.path.exists(fout):  return np.load(fout)

    # define the redshift and SFRH arrays
    z    = np.linspace(0.0, 7.0, bins)
    SFRH = np.zeros((realizations,bins), dtype=np.float32)

    # do a loop over all realizations and save results to file
    for i in range(realizations):
        fin = '%s/Results/SFRH/%s/%d/SFRH_0.00_10.00_10000.txt'%(root_in,sim,i)
        z_real, SFRH_real = np.loadtxt(fin, unpack=True)
        SFRH[i] = np.interp(z, z_real, SFRH_real)
    np.save(fout, SFRH)
    return SFRH

# This class creates the dataset 
class make_dataset():

    def __init__(self, mode, seed, realizations, root_in, bins, sim, root_out):

        # get the size and offset depending on the type of dataset
        if   mode=='train':  
            size, offset = int(realizations*0.70), int(realizations*0.00)
        elif mode=='valid':  
            size, offset = int(realizations*0.15), int(realizations*0.70)
        elif mode=='test':   
            size, offset = int(realizations*0.15), int(realizations*0.85)
        elif mode=='all':
            size, offset = int(realizations*1.00), int(realizations*0.00)
        else:    raise Exception('Wrong name!')

        # define size, input and output matrices
        self.size   = size
        self.input  = torch.zeros((size,6),    dtype=torch.float)
        self.output = torch.zeros((size,bins), dtype=torch.float)

        # read the value of the parameters and normalize them
        params     = np.loadtxt('%s/params_%s.txt'%(root_out,sim))
        min_params = np.min(params, axis=0)
        max_params = np.max(params, axis=0)
        params     = (params - min_params)/(max_params - min_params)

        # read the SFRH and normalize them
        SFRH = read_all_SFRH(root_in, sim, realizations, bins, root_out)
        SFRH[np.where(SFRH==0.0)] = 1e-12 #avoid points with SFRH=0
        SFRH = np.log10(SFRH)
        mean = np.mean(SFRH, axis=0, dtype=np.float64)
        std  = np.std(SFRH,  axis=0, dtype=np.float64)
        SFRH = (SFRH - mean)/std

        # randomly shuffle the cubes. Instead of 0 1 2 3...999 have a 
        # random permutation. E.g. 5 9 0 29...342
        np.random.seed(seed)
        indexes = np.arange(realizations) #only shuffle realizations, not rotations
        np.random.shuffle(indexes)
        indexes = indexes[offset:offset+size] #select indexes of mode

        # get the corresponding parameters and SFRH
        self.input  = torch.tensor(params[indexes], dtype=torch.float32)
        self.output = torch.tensor(SFRH[indexes],   dtype=torch.float32)

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return self.input[idx], self.output[idx]


# This routine creates a dataset loader
def create_dataset(mode, seed, realizations, root_in, bins, sim, batch_size, root_out):
    data_set = make_dataset(mode, seed, realizations, root_in, bins, sim, root_out)
    dataset_loader = DataLoader(dataset=data_set, batch_size=batch_size, shuffle=True)
    return dataset_loader



import numpy as np
import sys,os

################################### INPUT ###########################################
root = '/mnt/ceph/users/camels/Results/neural_nets/params_2_SFRH'
root_out = '/mnt/ceph/users/camels/Results/symbolic_regression'

f_SFRH   = '%s/SFRH_IllustrisTNG.npy'%root
f_params = '%s/params_IllustrisTNG.txt'%root

training_num = 700
testing_num  = 300
#####################################################################################

# read SFRH and params
SFRH   = np.load(f_SFRH);       SFRH = np.log10(SFRH)
params = np.loadtxt(f_params)
z      = np.linspace(0, 7, 100)

##### training set #####
data = np.zeros((training_num*100,6), dtype=np.float32)

# do a loop over the different realizations
count = 0
for i in range(training_num):
    
    # do a loop over the different redshfits
    for j in range(100):
        data[count,0] = z[j]+1.0
        data[count,1] = params[i,0]
        data[count,2] = params[i,1]
        data[count,3] = params[i,2]
        data[count,4] = params[i,4]
        data[count,5] = SFRH[i,j]
        count += 1

np.savetxt('%s/training_set.txt'%root_out, data)
########################

##### testing set #####
data = np.zeros((testing_num*100,6), dtype=np.float32)

# do a loop over the different realizations
count = 0
for i in range(training_num, training_num + testing_num):
    
    # do a loop over the different redshfits
    for j in range(100):
        data[count,0] = z[j]+1.0
        data[count,1] = params[i,0]
        data[count,2] = params[i,1]
        data[count,3] = params[i,2]
        data[count,4] = params[i,4]
        data[count,5] = SFRH[i,j]
        count += 1

np.savetxt('%s/test_set.txt'%root_out, data)
########################



import torch
import torch.nn as nn

def weights_init(m):
    """custom weights initialization
    """
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)


####################################################################################
####################################################################################
class autoencoder_64a(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64a, self).__init__()
        
        ##### first, contracting part #####
        # input: 1x64x64 ---------------> output: hiddenx32x32
        self.C1 = nn.Conv2d(1,         hidden, kernel_size=4, stride=2, padding=1, 
                            bias=True)
        self.B1 = nn.BatchNorm2d(hidden)
        # input: hiddenx32x32 ----------> output: 2*hiddenx16x16
        self.C2 = nn.Conv2d(hidden,   2*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B2 = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 --------> output: 4*hiddenx8x8
        self.C3 = nn.Conv2d(2*hidden, 4*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B3 = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ----------> output: 8*hiddenx4x4
        self.C4 = nn.Conv2d(4*hidden, 8*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B4 = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ----------> output: 500x1x1
        self.C5 = nn.Conv2d(8*hidden, BN_dim, kernel_size=6, stride=1, padding=1,
                            bias=True)
        self.B5 = nn.BatchNorm2d(BN_dim)

        ##### bottleneck #####
        self.FC1  = nn.Linear(500,     BN_dim)  #from C5 to bottleneck
        self.FC2  = nn.Linear(BN_dim,  500)     #from bottleneck to C6

        ##### second, expanding part #####
        # input: 500x1x1 ------------> output: 8*hiddenx4x4
        self.C6  = nn.ConvTranspose2d(BN_dim, 8*hidden, kernel_size=4, stride=1, 
                                      padding=0, bias=True)
        self.B6  = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ------------> output: 4*hiddenx8x8
        self.C7  = nn.ConvTranspose2d(8*hidden, 4*hidden, kernel_size=4, stride=2, 
                                      padding=1, bias=True)
        self.B7  = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ------------> output: 2*hiddenx16x16
        self.C8  = nn.ConvTranspose2d(4*hidden, 2*hidden, kernel_size=4, stride=2,
                                      padding=1, bias=True)
        self.B8  = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 ----------> output: hiddenx32x32
        self.C9 = nn.ConvTranspose2d(2*hidden, 1*hidden,  kernel_size=4, stride=2,
                                     padding=1, bias=True)
        self.B9 = nn.BatchNorm2d(hidden)
        # input: 1*hiddenx32x32 ----------> output: 1x64x64
        self.C10 = nn.ConvTranspose2d(hidden,  1,         kernel_size=4, stride=2,
                                      padding=1, bias=True)


        self.dropout   = nn.Dropout(p=0.5)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = self.LeakyReLU(self.C1(image))
        x = self.LeakyReLU(self.B2(self.C2(x)))
        x = self.LeakyReLU(self.B3(self.C3(x)))
        x = self.LeakyReLU(self.B4(self.C4(x)))
        #x = self.LeakyReLU(self.B5(self.C5(x)))
        x = self.LeakyReLU(self.C5(x))
        #inner_cnn = image.shape
        #x = x.view(inner_cnn[0],-1)

        #x = self.LeakyReLU(self.FC1(x))
        #x = self.LeakyReLU(self.FC2(x))

        #x = x.view(inner_cnn[0],inner_cnn[1],inner_cnn[2],inner_cnn[3])
        #x = self.LeakyReLU(self.B6(self.C6(x)))
        x = self.LeakyReLU(self.C6(x))
        x = self.LeakyReLU(self.B7(self.C7(x)))
        x = self.LeakyReLU(self.B8(self.C8(x)))
        x = self.LeakyReLU(self.B9(self.C9(x)))
        x = self.tanh(self.C10(x))

        return x
####################################################################################
####################################################################################

####################################################################################
####################################################################################
class autoencoder_64c(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64c, self).__init__()
        
        ##### first, contracting part #####
        # input: 1x64x64 ---------------> output: hiddenx32x32
        self.C1 = nn.Conv2d(1,         hidden, kernel_size=4, stride=2, padding=1, 
                            bias=True)
        self.B1 = nn.BatchNorm2d(hidden)
        # input: hiddenx32x32 ----------> output: 2*hiddenx16x16
        self.C2 = nn.Conv2d(hidden,   2*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B2 = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 --------> output: 4*hiddenx8x8
        self.C3 = nn.Conv2d(2*hidden, 4*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B3 = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ----------> output: 8*hiddenx4x4
        self.C4 = nn.Conv2d(4*hidden, 8*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B4 = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ----------> output: 500x1x1
        self.C5 = nn.Conv2d(8*hidden, BN_dim, kernel_size=6, stride=1, padding=1,
                            bias=True)
        self.B5 = nn.BatchNorm2d(BN_dim)

        ##### bottleneck #####
        self.FC1  = nn.Linear(500,     BN_dim)  #from C5 to bottleneck
        self.FC2  = nn.Linear(BN_dim,  500)     #from bottleneck to C6

        ##### second, expanding part #####
        # input: 500x1x1 ------------> output: 8*hiddenx4x4
        self.C6  = nn.ConvTranspose2d(BN_dim, 8*hidden, kernel_size=4, stride=1, 
                                      padding=0, bias=True)
        self.B6  = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ------------> output: 4*hiddenx8x8
        self.C7  = nn.ConvTranspose2d(8*hidden, 4*hidden, kernel_size=4, stride=2, 
                                      padding=1, bias=True)
        self.B7  = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ------------> output: 2*hiddenx16x16
        self.C8  = nn.ConvTranspose2d(4*hidden, 2*hidden, kernel_size=4, stride=2,
                                      padding=1, bias=True)
        self.B8  = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 ----------> output: hiddenx32x32
        self.C9 = nn.ConvTranspose2d(2*hidden, 1*hidden,  kernel_size=4, stride=2,
                                     padding=1, bias=True)
        self.B9 = nn.BatchNorm2d(hidden)
        # input: 1*hiddenx32x32 ----------> output: 1x64x64
        self.C10 = nn.ConvTranspose2d(hidden,  1,         kernel_size=4, stride=2,
                                      padding=1, bias=True)


        self.dropout   = nn.Dropout(p=0.5)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = self.LeakyReLU(self.C1(image))
        x = self.LeakyReLU(self.C2(x))
        x = self.LeakyReLU(self.C3(x))
        x = self.LeakyReLU(self.C4(x))
        x = self.LeakyReLU(self.C5(x))
        #inner_cnn = x.shape
        #x = x.view(inner_cnn[0],-1)

        #x = self.LeakyReLU(self.FC1(x))
        #x = self.LeakyReLU(self.FC2(x))

        #x = x.view(inner_cnn[0],inner_cnn[1],inner_cnn[2],inner_cnn[3])
        #x = self.LeakyReLU(self.B6(self.C6(x)))
        x = self.LeakyReLU(self.C6(x))
        x = self.LeakyReLU(self.C7(x))
        x = self.LeakyReLU(self.C8(x))
        x = self.LeakyReLU(self.C9(x))
        x = self.tanh(self.C10(x))

        return x
####################################################################################
####################################################################################

####################################################################################
####################################################################################
class autoencoder_64d(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64d, self).__init__()
        
        ##### first, contracting part #####
        # input: 1x64x64 ---------------> output: hiddenx32x32
        self.C1 = nn.Conv2d(1,         hidden, kernel_size=4, stride=2, padding=1, 
                            bias=True)
        self.B1 = nn.BatchNorm2d(hidden)
        # input: hiddenx32x32 ----------> output: 2*hiddenx16x16
        self.C2 = nn.Conv2d(hidden,   2*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B2 = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 --------> output: 4*hiddenx8x8
        self.C3 = nn.Conv2d(2*hidden, 4*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B3 = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ----------> output: 8*hiddenx4x4
        self.C4 = nn.Conv2d(4*hidden, 8*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B4 = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ----------> output: 1000x1x1
        self.C5 = nn.Conv2d(8*hidden, 1000, kernel_size=6, stride=1, padding=1,
                            bias=True)
        self.B5 = nn.BatchNorm2d(BN_dim)

        ##### bottleneck #####
        self.FC1  = nn.Linear(1000,    500)   
        self.FC2  = nn.Linear(500,     250)   
        self.FC3  = nn.Linear(250,     BN_dim)   
        self.FC4  = nn.Linear(BN_dim,  250)   
        self.FC5  = nn.Linear(250,     500)   
        self.FC6  = nn.Linear(500,     1000)   

        ##### second, expanding part #####
        # input: 1000x1x1 ------------> output: 8*hiddenx4x4
        self.C6  = nn.ConvTranspose2d(1000, 8*hidden, kernel_size=4, stride=1, 
                                      padding=0, bias=True)
        self.B6  = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ------------> output: 4*hiddenx8x8
        self.C7  = nn.ConvTranspose2d(8*hidden, 4*hidden, kernel_size=4, stride=2, 
                                      padding=1, bias=True)
        self.B7  = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ------------> output: 2*hiddenx16x16
        self.C8  = nn.ConvTranspose2d(4*hidden, 2*hidden, kernel_size=4, stride=2,
                                      padding=1, bias=True)
        self.B8  = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 ----------> output: hiddenx32x32
        self.C9 = nn.ConvTranspose2d(2*hidden, 1*hidden,  kernel_size=4, stride=2,
                                     padding=1, bias=True)
        self.B9 = nn.BatchNorm2d(hidden)
        # input: 1*hiddenx32x32 ----------> output: 1x64x64
        self.C10 = nn.ConvTranspose2d(hidden,  1,         kernel_size=4, stride=2,
                                      padding=1, bias=True)


        self.dropout   = nn.Dropout(p=0.5)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = self.LeakyReLU(self.C1(image))
        x = self.LeakyReLU(self.B2(self.C2(x)))
        x = self.LeakyReLU(self.B3(self.C3(x)))
        x = self.LeakyReLU(self.B4(self.C4(x)))
        x = self.LeakyReLU(self.C5(x))
        inner_cnn = x.shape
        x = x.view(inner_cnn[0],-1)

        x = self.LeakyReLU(self.FC1(x))
        x = self.LeakyReLU(self.FC2(x))
        x = self.LeakyReLU(self.FC3(x))
        x = self.LeakyReLU(self.FC4(x))
        x = self.LeakyReLU(self.FC5(x))
        x = self.LeakyReLU(self.FC6(x))

        x = x.view(inner_cnn[0],inner_cnn[1],inner_cnn[2],inner_cnn[3])
        x = self.LeakyReLU(self.C6(x))
        x = self.LeakyReLU(self.B7(self.C7(x)))
        x = self.LeakyReLU(self.B8(self.C8(x)))
        x = self.LeakyReLU(self.B9(self.C9(x)))
        x = self.tanh(self.C10(x))

        return x
####################################################################################
####################################################################################

####################################################################################
####################################################################################
class autoencoder_64e(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64e, self).__init__()
        
        ##### first, contracting part #####
        # input: 1x64x64 ---------------> output: hiddenx16x16
        self.C1 = nn.Conv2d(1,         hidden, kernel_size=6, stride=4, padding=1, 
                            bias=True)
        self.B1 = nn.BatchNorm2d(hidden)
        # input: hiddenx16x16 ----------> output: 2*hiddenx4x4
        self.C2 = nn.Conv2d(hidden,   2*hidden, kernel_size=6, stride=4, padding=1,
                            bias=True)
        self.B2 = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx4x4 --------> output: BN_dimx1x1
        self.C3 = nn.Conv2d(2*hidden, BN_dim, kernel_size=6, stride=4, padding=1,
                            bias=True)
        self.B3 = nn.BatchNorm2d(4*hidden)

        ##### bottleneck #####
        self.FC1  = nn.Linear(500,     BN_dim)  #from C5 to bottleneck
        self.FC2  = nn.Linear(BN_dim,  500)     #from bottleneck to C6

        ##### second, expanding part #####
        # input: BN_dimx1x1 ------------> output: 2*hiddenx4x4
        self.C6  = nn.ConvTranspose2d(BN_dim, 2*hidden, kernel_size=4, stride=1, 
                                      padding=0, bias=True)
        self.B6  = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx4x4 ------------> output: hiddenx16x16
        self.C7  = nn.ConvTranspose2d(2*hidden, 1*hidden, kernel_size=6, stride=4, 
                                      padding=1, bias=True)
        self.B7  = nn.BatchNorm2d(1*hidden)
        # input: 2*hiddenx16x16 ------------> output: 1x64x64
        self.C8  = nn.ConvTranspose2d(1*hidden, 1,        kernel_size=6, stride=4,
                                      padding=1, bias=True)

        self.dropout   = nn.Dropout(p=0.5)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = self.LeakyReLU(self.C1(image))
        x = self.LeakyReLU(self.B2(self.C2(x)))
        x = self.LeakyReLU(self.C3(x))

        #inner_cnn = x.shape
        #x = x.view(inner_cnn[0],-1)

        #x = self.LeakyReLU(self.FC1(x))
        #x = self.LeakyReLU(self.FC2(x))

        #x = x.view(inner_cnn[0],inner_cnn[1],inner_cnn[2],inner_cnn[3])
        #x = self.LeakyReLU(self.B6(self.C6(x)))

        x = self.LeakyReLU(self.C6(x))
        x = self.LeakyReLU(self.B7(self.C7(x)))
        x = self.tanh(self.C8(x))

        return x
####################################################################################
####################################################################################

####################################################################################
####################################################################################
class autoencoder_64f(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64f, self).__init__()
        
        ##### first, contracting part #####
        # input: 1x64x64 ---------------> output: hiddenx16x16
        self.C1 = nn.Conv2d(1,         hidden, kernel_size=6, stride=4, padding=1, 
                            bias=True)
        self.B1 = nn.BatchNorm2d(hidden)
        # input: hiddenx16x16 ----------> output: 2*hiddenx4x4
        self.C2 = nn.Conv2d(hidden,   2*hidden, kernel_size=6, stride=4, padding=1,
                            bias=True)
        self.B2 = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx4x4 --------> output: BN_dimx1x1
        self.C3 = nn.Conv2d(2*hidden, BN_dim, kernel_size=6, stride=4, padding=1,
                            bias=True)
        self.B3 = nn.BatchNorm2d(BN_dim)

        ##### bottleneck #####
        self.FC1  = nn.Linear(500,     BN_dim)  #from C5 to bottleneck
        self.FC2  = nn.Linear(BN_dim,  500)     #from bottleneck to C6

        ##### second, expanding part #####
        # input: BN_dimx1x1 ------------> output: 2*hiddenx4x4
        self.C6  = nn.ConvTranspose2d(BN_dim, 2*hidden, kernel_size=4, stride=1, 
                                      padding=0, bias=True)
        self.B6  = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx4x4 ------------> output: hiddenx16x16
        self.C7  = nn.ConvTranspose2d(2*hidden, 1*hidden, kernel_size=6, stride=4, 
                                      padding=1, bias=True)
        self.B7  = nn.BatchNorm2d(1*hidden)
        # input: 2*hiddenx16x16 ------------> output: 1x64x64
        self.C8  = nn.ConvTranspose2d(1*hidden, 1,        kernel_size=6, stride=4,
                                      padding=1, bias=True)

        self.dropout   = nn.Dropout(p=0.5)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = self.LeakyReLU(self.B1(self.C1(image)))
        x = self.LeakyReLU(self.B2(self.C2(x)))
        x = self.LeakyReLU(self.B3(self.C3(x)))

        #inner_cnn = x.shape
        #x = x.view(inner_cnn[0],-1)

        #x = self.LeakyReLU(self.FC1(x))
        #x = self.LeakyReLU(self.FC2(x))

        #x = x.view(inner_cnn[0],inner_cnn[1],inner_cnn[2],inner_cnn[3])
        #x = self.LeakyReLU(self.B6(self.C6(x)))

        x = self.LeakyReLU(self.B6(self.C6(x)))
        x = self.LeakyReLU(self.B7(self.C7(x)))
        x = self.tanh(self.C8(x))

        return x
####################################################################################
####################################################################################

####################################################################################
####################################################################################
class autoencoder_64g(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64g, self).__init__()
        
        ##### first, contracting part #####
        # input: 1x64x64 ---------------> output: hiddenx16x16
        self.C1 = nn.Conv2d(1,         hidden, kernel_size=6, stride=4, padding=1, 
                            bias=True)
        self.B1 = nn.BatchNorm2d(hidden)
        # input: hiddenx16x16 ----------> output: 2*hiddenx4x4
        self.C2 = nn.Conv2d(hidden,   2*hidden, kernel_size=6, stride=4, padding=1,
                            bias=True)
        self.B2 = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx4x4 --------> output: BN_dimx1x1
        self.C3 = nn.Conv2d(2*hidden, BN_dim, kernel_size=6, stride=4, padding=1,
                            bias=True)
        self.B3 = nn.BatchNorm2d(BN_dim)

        ##### bottleneck #####
        self.FC1  = nn.Linear(500,     BN_dim)  #from C5 to bottleneck
        self.FC2  = nn.Linear(BN_dim,  500)     #from bottleneck to C6

        ##### second, expanding part #####
        # input: BN_dimx1x1 ------------> output: 2*hiddenx4x4
        self.C6  = nn.ConvTranspose2d(BN_dim, 2*hidden, kernel_size=4, stride=1, 
                                      padding=0, bias=True)
        self.B6  = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx4x4 ------------> output: hiddenx16x16
        self.C7  = nn.ConvTranspose2d(2*hidden, 1*hidden, kernel_size=6, stride=4, 
                                      padding=1, bias=True)
        self.B7  = nn.BatchNorm2d(1*hidden)
        # input: 2*hiddenx16x16 ------------> output: 1x64x64
        self.C8  = nn.ConvTranspose2d(1*hidden, 1,        kernel_size=6, stride=4,
                                      padding=1, bias=True)

        self.dropout   = nn.Dropout(p=0.5)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = self.LeakyReLU(self.C1(image))
        x = self.LeakyReLU(self.C2(x))
        x = self.LeakyReLU(self.C3(x))

        x = self.LeakyReLU(self.C6(x))
        x = self.LeakyReLU(self.C7(x))
        x = self.tanh(self.C8(x))

        return x
####################################################################################
####################################################################################

####################################################################################
####################################################################################
class autoencoder_64h(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64h, self).__init__()
        
        ##### first, contracting part #####
        # input: 1x64x64 ---------------> output: hiddenx32x32
        self.C1 = nn.Conv2d(1,         hidden, kernel_size=4, stride=2, padding=1, 
                            bias=True)
        self.B1 = nn.BatchNorm2d(hidden)
        # input: hiddenx32x32 ----------> output: 2*hiddenx16x16
        self.C2 = nn.Conv2d(hidden,   2*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B2 = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 --------> output: 4*hiddenx8x8
        self.C3 = nn.Conv2d(2*hidden, 4*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B3 = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ----------> output: 8*hiddenx4x4
        self.C4 = nn.Conv2d(4*hidden, 8*hidden, kernel_size=4, stride=2, padding=1,
                            bias=True)
        self.B4 = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ----------> output: 1000x1x1
        self.C5 = nn.Conv2d(8*hidden, 1000, kernel_size=6, stride=1, padding=1,
                            bias=True)
        self.B5 = nn.BatchNorm2d(1000)

        ##### bottleneck #####
        self.FC1  = nn.Linear(1000,    BN_dim)  #from C5 to bottleneck
        self.FC2  = nn.Linear(BN_dim,  1000)     #from bottleneck to C6

        ##### second, expanding part #####
        # input: 1000x1x1 ------------> output: 8*hiddenx4x4
        self.C6  = nn.ConvTranspose2d(1000, 8*hidden, kernel_size=4, stride=1, 
                                      padding=0, bias=True)
        self.B6  = nn.BatchNorm2d(8*hidden)
        # input: 8*hiddenx4x4 ------------> output: 4*hiddenx8x8
        self.C7  = nn.ConvTranspose2d(8*hidden, 4*hidden, kernel_size=4, stride=2, 
                                      padding=1, bias=True)
        self.B7  = nn.BatchNorm2d(4*hidden)
        # input: 4*hiddenx8x8 ------------> output: 2*hiddenx16x16
        self.C8  = nn.ConvTranspose2d(4*hidden, 2*hidden, kernel_size=4, stride=2,
                                      padding=1, bias=True)
        self.B8  = nn.BatchNorm2d(2*hidden)
        # input: 2*hiddenx16x16 ----------> output: hiddenx32x32
        self.C9 = nn.ConvTranspose2d(2*hidden, 1*hidden,  kernel_size=4, stride=2,
                                     padding=1, bias=True)
        self.B9 = nn.BatchNorm2d(hidden)
        # input: 1*hiddenx32x32 ----------> output: 1x64x64
        self.C10 = nn.ConvTranspose2d(hidden,  1,         kernel_size=4, stride=2,
                                      padding=1, bias=True)


        self.dropout   = nn.Dropout(p=0.5)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = self.LeakyReLU(self.C1(image))
        x = self.LeakyReLU(self.B2(self.C2(x)))
        x = self.LeakyReLU(self.B3(self.C3(x)))
        x = self.LeakyReLU(self.B4(self.C4(x)))
        x = self.LeakyReLU(self.C5(x))
        dims = x.shape
        x = x.view(dims[0],-1)

        x = self.LeakyReLU(self.FC1(x))
        x = self.LeakyReLU(self.FC2(x))

        x = x.view(dims[0],dims[1],dims[2],dims[3])
        x = self.LeakyReLU(self.C6(x))
        x = self.LeakyReLU(self.B7(self.C7(x)))
        x = self.LeakyReLU(self.B8(self.C8(x)))
        x = self.LeakyReLU(self.B9(self.C9(x)))
        x = self.tanh(self.C10(x))

        return x
####################################################################################
####################################################################################

####################################################################################
####################################################################################
class autoencoder_64b(nn.Module):
    def __init__(self, BN_dim, hidden):
        super(autoencoder_64b, self).__init__()
        
        ##### first, contracting part #####
        self.FC1 = nn.Linear(64*64, 1000)
        self.FC2 = nn.Linear(1000,  500)
        self.FC3 = nn.Linear(500,   250)
        self.FC4 = nn.Linear(250,   BN_dim)
        
        ##### second, expanding part #####
        self.FC5 = nn.Linear(BN_dim, 250)
        self.FC6 = nn.Linear(250,    500)
        self.FC7 = nn.Linear(500,    1000)
        self.FC8 = nn.Linear(1000,   64*64)


        self.dropout   = nn.Dropout(p=0.3)
        self.ReLU      = nn.ReLU()
        self.LeakyReLU = nn.LeakyReLU(0.2)
        self.tanh      = nn.Tanh()

        """
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 1)
            elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight)
        """


    def forward(self, image):
        x = image.view(image.size(0),-1)
        x = self.LeakyReLU(self.FC1(x))
        x = self.LeakyReLU(self.FC2(x))
        x = self.LeakyReLU(self.FC3(x))
        x = self.LeakyReLU(self.FC4(x))
        x = self.LeakyReLU(self.FC5(x))
        x = self.LeakyReLU(self.FC6(x))
        x = self.LeakyReLU(self.FC7(x))
        x = self.tanh(self.FC8(x))
        x = x.view(image.size(0),image.size(1),image.size(2),image.size(3))

        return x
####################################################################################
####################################################################################




# ConvTranspose2d(channels_in, channels_out, kernel, stride, padding)
class Generator_64(nn.Module):
    def __init__(self, Z_DIM, G_HIDDEN):
        super(Generator_64, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 100x1x1 ----> output: 512x4x4)
            nn.ConvTranspose2d(Z_DIM, G_HIDDEN * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 8),
            nn.ReLU(True),
            # 2nd layer (input: 512x4x4 ----> output: 256x8x8)
            nn.ConvTranspose2d(G_HIDDEN * 8, G_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 4),
            nn.ReLU(True),
            # 3rd layer (input: 256x8x8 ----> output: 128x16x16)
            nn.ConvTranspose2d(G_HIDDEN * 4, G_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 2),
            nn.ReLU(True),
            # 4th layer (input: 128x16x16 ----> output: 64x32x32)
            nn.ConvTranspose2d(G_HIDDEN * 2, G_HIDDEN, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN),
            nn.ReLU(True),
            # output layer (input: 64x32x32 ----> 1x64x64)
            nn.ConvTranspose2d(G_HIDDEN, 1, 4, 2, 1, bias=False),
            nn.Tanh()
        )

    def forward(self, input):
        return self.main(input)


#Conv2d(channels_in, channels_out, kernel, stride, padding)
class Discriminator_64(nn.Module):
    def __init__(self, D_HIDDEN):
        super(Discriminator_64, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 1x64x64 ----> output: 64x32x32)
            nn.Conv2d(1, D_HIDDEN, 4, 2, 1, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # 2nd layer (input: 64x32x32 ----> output: 128x16x16)
            nn.Conv2d(D_HIDDEN, D_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # 3rd layer (input: 128x16x16 ----> output: 256x8x8)
            nn.Conv2d(D_HIDDEN * 2, D_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # 4th layer (input: 256x8x8 ----> output: 512x4x4)
            nn.Conv2d(D_HIDDEN * 4, D_HIDDEN * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # output layer (input: 512x4x4 ----> output: 1x1x1)
            nn.Conv2d(D_HIDDEN * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid() #Returns a number from 0 to 1: probability
        )

    def forward(self, input):
        return self.main(input).view(-1, 1).squeeze(1)







# ConvTranspose2d(channels_in, channels_out, kernel, stride, padding)
class Generator_128(nn.Module):
    def __init__(self, Z_DIM, G_HIDDEN):
        super(Generator_128, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 100x1x1 ----> output: 512x8x8)
            nn.ConvTranspose2d(Z_DIM, G_HIDDEN * 8, 8, 1, 0, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 8),
            nn.ReLU(True),
            # 2nd layer (input: 512x8x8 ----> output: 256x16x16)
            nn.ConvTranspose2d(G_HIDDEN * 8, G_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 4),
            nn.ReLU(True),
            # 3rd layer (input: 256x16x16 ----> output: 128x32x32)
            nn.ConvTranspose2d(G_HIDDEN * 4, G_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 2),
            nn.ReLU(True),
            # 4th layer (input: 128x32x32 ----> output: 64x64x64)
            nn.ConvTranspose2d(G_HIDDEN * 2, G_HIDDEN, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN),
            nn.ReLU(True),
            # output layer (input: 64x64x64 ----> 1x128x128)
            nn.ConvTranspose2d(G_HIDDEN, 1, 4, 2, 1, bias=False),
        )

    def forward(self, input):
        return self.main(input)


#Conv2d(channels_in, channels_out, kernel, stride, padding)
class Discriminator_128(nn.Module):
    def __init__(self, D_HIDDEN):
        super(Discriminator_128, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 1x128x128 ----> output: 64x32x32)
            nn.Conv2d(1, D_HIDDEN, 4, 4, 0, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # 2nd layer (input: 64x32x32 ----> output: 128x16x16)
            nn.Conv2d(D_HIDDEN, D_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # 3rd layer (input: 128x16x16 ----> output: 256x8x8)
            nn.Conv2d(D_HIDDEN * 2, D_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # 4th layer (input: 256x8x8 ----> output: 512x4x4)
            nn.Conv2d(D_HIDDEN * 4, D_HIDDEN * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # output layer (input: 512x4x4 ----> output: 1x1x1)
            nn.Conv2d(D_HIDDEN * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid() #Returns a number from 0 to 1: probability
        )

    def forward(self, input):
        return self.main(input).view(-1, 1).squeeze(1)




from pylab import *
import numpy as np
from matplotlib.ticker import ScalarFormatter
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1.inset_locator import mark_inset
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import LogNorm
from matplotlib.patches import Ellipse
rcParams["mathtext.fontset"]='cm'

############################### figure ###########################
#fig=figure(figsize=(15,10))     #give dimensions to the figure
##################################################################

################################ INPUT #######################################
#axes range
##############################################################################

############################ subplots ############################
#gs = gridspec.GridSpec(2,1,height_ratios=[5,2])
#ax1=plt.subplot(gs[0])
#ax2=plt.subplot(gs[1])

#make a subplot at a given position and with some given dimensions
#ax2=axes([0.4,0.55,0.25,0.1])

#gs.update(hspace=0.0,wspace=0.4,bottom=0.6,top=1.05)
#subplots_adjust(left=None, bottom=None, right=None, top=None,
#                wspace=0.5, hspace=0.5)

#set minor ticks
#ax1.xaxis.set_minor_locator(AutoMinorLocator(4))
#ax1.yaxis.set_minor_locator(AutoMinorLocator(4))


#ax1.xaxis.set_major_formatter( NullFormatter() )   #unset x label 
#ax1.yaxis.set_major_formatter( NullFormatter() )   #unset y label

#ax1.set_xticks([])
#ax1.set_yticks([])


#ax1.get_yaxis().set_label_coords(-0.2,0.5)  #align y-axis for multiple plots
##################################################################

##################### special behaviour stuff ####################
#to show error missing error bars in log scale
#ax1.set_yscale('log',nonposy='clip')  #set log scale for the y-axis

#set the x-axis in %f format instead of %e
#ax1.xaxis.set_major_formatter(ScalarFormatter()) 

#set size of ticks
#ax1.tick_params(axis='both', which='major', labelsize=10)
#ax1.tick_params(axis='both', which='minor', labelsize=8)

#set the position of the ylabel 
#ax1.yaxis.set_label_coords(-0.2, 0.4)

#set yticks in scientific notation
#ax1.ticklabel_format(axis='y',style='sci',scilimits=(1,4))

#set the x-axis in %f format instead of %e
#formatter = matplotlib.ticker.FormatStrFormatter('$%.2e$') 
#ax1.yaxis.set_major_formatter(formatter) 

#add two legends in the same plot
#ax5 = ax1.twinx()
#ax5.yaxis.set_major_formatter( NullFormatter() )   #unset y label 
#ax5.legend([p1,p2],['0.0 eV','0.3 eV'],loc=3,prop={'size':14},ncol=1)

#set points to show in the yaxis
#ax1.set_yticks([0,1,2])

#highlight a zoomed region
#mark_inset(ax1, ax2, loc1=2, loc2=4, fc="none",edgecolor='purple')
##################################################################

############################ plot type ###########################
#standard plot
#p1,=ax1.plot(x,y,linestyle='-',marker='None')

#error bar plot with the minimum and maximum values of the error bar interval
#p1=ax1.errorbar(r,xi,yerr=[delta_xi_min,delta_xi_max],lw=1,fmt='o',ms=2,
#               elinewidth=1,capsize=5,linestyle='-') 

#filled area
#p1=ax1.fill_between([x_min,x_max],[1.02,1.02],[0.98,0.98],color='k',alpha=0.2)

#hatch area
#ax1.fill([x_min,x_min,x_max,x_max],[y_min,3.0,3.0,y_min],#color='k',
#         hatch='X',fill=False,alpha=0.5)

#scatter plot
#p1=ax1.scatter(k1,Pk1,c='b',edgecolor='none',s=8,marker='*')

#plot with markers
#pl4,=ax1.plot(ke3,Pk3/Pke3,marker='.',markevery=2,c='r',linestyle='None')

#set size of dashed lines
#ax.plot([0, 1], [0, 1], linestyle='--', dashes=(5, 1)) #length of 5, space of 1

#image plot
#cax = ax1.imshow(densities,cmap=get_cmap('jet'),origin='lower',
#           extent=[x_min, x_max, y_min, y_max],
#           #vmin=min_density,vmax=max_density)
#           norm = LogNorm(vmin=min_density,vmax=max_density))
#cbar = fig.colorbar(cax, ax2, ax=ax1, ticks=[-1, 0, 1]) #in ax2 colorbar of ax1
#cbar.set_label(r"$M_{\rm CSF}\/[h^{-1}M_\odot]$",fontsize=14,labelpad=-50)
#cbar.ax.tick_params(labelsize=10)  #to change size of ticks

#make a polygon
#polygon = Rectangle((0.4,50.0), 20.0, 20.0, edgecolor='purple',lw=0.5,
#                    fill=False)
#ax1.add_artist(polygon)
####################################################################

x_min, x_max = -0.1, 7.1
y_min, y_max = 4e-3, 1.0

fig = figure()
ax1 = fig.add_subplot(111) 

#ax1.set_xscale('log')
ax1.set_yscale('log')

ax1.set_xlim([x_min,x_max])
#ax1.set_ylim([y_min,y_max])

ax1.set_xlabel(r'$z$',fontsize=18)
ax1.set_ylabel(r'${\rm SFRH}\,\,[M_\odot{\rm yr}^{-1}{\rm Mpc}^{-3}]$',fontsize=18)

f_out = 'SFRH_NN_TNG.pdf'

f1 = '../../../../Results/neural_nets/params_2_SFRH/SFRH_ML.txt'

data = np.loadtxt(f1)
z = data[:,0]
SFRH_pred = data[:,1:151]
SFRH_true = data[:,151:301]

#i,j,k,l = 0,1,7,13
i,j,k,l = 20,35,65,120

p1,=ax1.plot(z,SFRH_pred[:,i],linestyle='-',marker='None',c='r')
p2,=ax1.plot(z,SFRH_true[:,i],linestyle='--',marker='None',c='r')

ax1.plot(z,SFRH_pred[:,j],linestyle='-',marker='None',c='g')
ax1.plot(z,SFRH_true[:,j],linestyle='--',marker='None',c='g')

ax1.plot(z,SFRH_pred[:,k],linestyle='-',marker='None',c='b')
ax1.plot(z,SFRH_true[:,k],linestyle='--',marker='None',c='b')

ax1.plot(z,SFRH_pred[:,l],linestyle='-',marker='None',c='purple')
ax1.plot(z,SFRH_true[:,l],linestyle='--',marker='None',c='purple')

#place a label in the plot
#ax1.text(0.2,0.1, r"$z=4.0$", fontsize=22, color='k')

#legend
ax1.legend([p1,p2],
           [r"${\rm Neural\,\,network}$",
            r"${\rm IllustrisTNG}$"],
           loc=0,prop={'size':15},ncol=1,frameon=True)
            
            #columnspacing=2,labelspacing=2)




#ax1.set_title(r'$\sum m_\nu=0.0\/{\rm eV}$',position=(0.5,1.02),size=18)
#title('About as simple as it gets, folks')
#suptitle('About as simple as it gets, folks')  #for title with several panels
#grid(True)
#show()
savefig(f_out, bbox_inches='tight')
close(fig)










###############################################################################
#some useful colors:

#'darkseagreen'
#'yellow'
#"hotpink"
#"gold
#"fuchsia"
#"lime"
#"brown"
#"silver"
#"cyan"
#"dodgerblue"
#"darkviolet"
#"magenta"
#"deepskyblue"
#"orchid"
#"aqua"
#"darkorange"
#"coral"
#"lightgreen"
#"salmon"
#"bisque"



import numpy as np
import sys,os,h5py

root        = '/simons/scratch/sgenel/OUTPUT_RUNS/CosmologicalBoxes/Hydro_DL/'
realization = 1505
snapnum     = 33


fin = '%s/%d/fof_subhalo_tab_%03d.hdf5'%(root,realization,snapnum)

f = h5py.File(fin, 'r')
mass = f['Group/GroupMass'][:]*1e10 #Msun/h
f.close()

print('%.3e < Mass < %.3e'%(np.min(mass), np.max(mass)))



# This script creates 2D images/fields from different fields of the simulations
from mpi4py import MPI
import numpy as np
import sys,os,h5py
import MAS_library as MASL
sys.path.append('../analysis')
import camel_library as CL

###### MPI DEFINITIONS ###### 
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

#################################### INPUT ##########################################
# data parameters
root_in       = '/mnt/ceph/users/camels'
root_out      = '/mnt/ceph/users/camels/Results/images_1P'
#TREECOOL_file = '/mnt/ceph/users/camels/Software/TREECOOL'
sim           = 'SIMBA'
snapnum       = 33
realizations  = 66

# images parameters
grid         = 250
splits       = 5

# parameters of the density field routine
periodic     = True
verbose      = False
x_min, y_min = 0.0, 0.0
#####################################################################################

# find the numbers that each cpu will work with
numbers = np.where(np.arange(realizations)%nprocs==myrank)[0]

# define the matrix hosting all the maps
maps_T_loc  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_T_tot  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Z_loc  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Z_tot  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_P_loc  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_P_tot  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Mg_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Mg_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Mc_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Mc_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Ms_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Ms_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Vg_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_Vg_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_HI_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_HI_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_ne_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
maps_ne_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)

# do a loop over all realizations
for i in numbers:

    if i in [16,27,38,49,60]:  index = 1505
    else:                      index = 1500+i

    if myrank==0:  print(i, end=' ', flush=True)
    
    # read the gas positions, radii and temperature of gas particles
    snapshot = '%s/Sims/%s/%d/snap_%03d.hdf5'%(root_in,sim,index,snapnum)
    if not(os.path.exists(snapshot)):  raise Exception('%s not found'%snapshot)
    f = h5py.File(snapshot, 'r')
    BoxSize = f['Header'].attrs[u'BoxSize']/1e3 #Mpc/h
    redshift = f['Header'].attrs[u'Redshift'];  
    Masses  = f['Header'].attrs[u'MassTable'][:]*1e10
    pos_g   = f['PartType0/Coordinates'][:]/1e3;  pos_g = pos_g.astype(np.float32)#Mpc/h
    pos_c   = f['PartType1/Coordinates'][:]/1e3;  pos_c = pos_c.astype(np.float32)#Mpc/h
    pos_s   = f['PartType4/Coordinates'][:]/1e3;  pos_s = pos_s.astype(np.float32)#Mpc/h
    vel_g   = f['/PartType0/Velocities'][:]/np.sqrt(1.0+redshift)
    vel_g   = np.sqrt(vel_g[:,0]**2 + vel_g[:,1]**2 + vel_g[:,2]**2)
    Mg      = f['PartType0/Masses'][:]*1e10     #Msun/h
    Ms      = f['PartType4/Masses'][:]*1e10     #Msun/h
    if sim=='IllustrisTNG':
        Mc  = np.ones(pos_c.shape[0], dtype=np.float32)*Masses[1] #Msun/h
        Rg  = f['PartType0/SubfindHsml'][:]/1e3  #Mpc/h
        Rc  = f['PartType1/SubfindHsml'][:]/1e3  #Mpc/h
        Rs  = f['PartType4/SubfindHsml'][:]/1e3  #Mpc/h
        Z   = f['/PartType0/GFM_Metallicity'][:] #metallicity
        HI  = CL.HI_mass(snapshot, TREECOOL_file)
    else:
        Mc  = f['PartType1/Masses'][:]*1e10       #Msun/h
        Rg  = f['PartType0/AGS-Softening'][:]/1e3 #Mpc/h
        Rc  = f['PartType1/AGS-Softening'][:]/1e3 #Mpc/h
        Rs  = f['PartType4/AGS-Softening'][:]/1e3 #Mpc/h
        Z   = f['/PartType0/Metallicity'][:,0] #metallicity
        HI  = f['/PartType0/NeutralHydrogenAbundance'][:]  #HI/H
    f.close()
    T  = CL.temperature(snapshot)      #K
    P  = CL.pressure(snapshot)         #Msun*(km/s)^2/kpc^3
    ne = CL.electron_density(snapshot) #1e20 electrons*h^2/Mpc^3

    # do a loop over the three different axes
    for axis in [0,1,2]:

        axis_x, axis_y = (axis+1)%3, (axis+2)%3 

        # do a loop over the different slices of each axis
        for j in range(splits):
            
            # get the number of the map; this is important for MPI
            num = i*3*splits + axis*splits + j

            # find the range in the slice
            minimum, maximum = j*BoxSize/splits, (j+1)*BoxSize/splits

            # select the particles in the considered slice
            indexes_g = np.where((pos_g[:,axis]>=minimum) & (pos_g[:,axis]<maximum))[0] 
            pos_g_    = pos_g[indexes_g]
            vel_g_    = vel_g[indexes_g]
            Rg_       = Rg[indexes_g]
            Mg_       = Mg[indexes_g]
            T_        = T[indexes_g]
            Z_        = Z[indexes_g]
            HI_       = HI[indexes_g]
            P_        = P[indexes_g]
            ne_       = ne[indexes_g]
            indexes_c = np.where((pos_c[:,axis]>=minimum) & (pos_c[:,axis]<maximum))[0] 
            pos_c_    = pos_c[indexes_c]
            Mc_       = Mc[indexes_c]
            Rc_       = Rc[indexes_c]
            indexes_s = np.where((pos_s[:,axis]>=minimum) & (pos_s[:,axis]<maximum))[0] 
            pos_s_    = pos_s[indexes_s]
            Ms_       = Ms[indexes_s]
            Rs_       = Rs[indexes_s]
            
            # project gas mass*temperatures into a 2D map
            TM = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(TM, pos_g_, T_*Mg_, Rg_, x_min, y_min, 
                               axis_x, axis_y, BoxSize, periodic, verbose)
            
            # project gas mass*metallicity into a 2D map
            ZM = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(ZM, pos_g_, Z_*Mg_, Rg_, x_min, y_min, 
                               axis_x, axis_y, BoxSize, periodic, verbose)
            
            # project gas velocity*mass into a 2D map
            VM = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(VM, pos_g_, vel_g_*Mg_, Rg_, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose)
            
            # project gas pressure*mass into a 2D map
            PM = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(PM, pos_g_, P_*Mg_, Rg_, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose)

            # project gas mass into a 2D map
            Mgas = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(Mgas, pos_g_, Mg_, Rg_, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose)

            # project cdm mass into a 2D map
            Mcdm = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(Mcdm, pos_c_, Mc_, Rc_, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose)

            # project stellar mass into a 2D map
            Mstar = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(Mstar, pos_s_, Ms_, Rs_, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose)
                
            # project HI mass into a 2D map
            HIgas = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(HIgas, pos_g_, HI_, Rg_, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose)

            # project electron density into a 2D map
            negas = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(negas, pos_g_, ne_, Rg_, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose)

            maps_T_loc[num]  = TM/Mgas
            maps_Z_loc[num]  = ZM/Mgas
            maps_P_loc[num]  = PM/Mgas
            maps_Vg_loc[num] = VM/Mgas
            maps_Mg_loc[num] = Mgas
            maps_Mc_loc[num] = Mcdm
            maps_Ms_loc[num] = Mstar
            maps_HI_loc[num] = HIgas
            maps_ne_loc[num] = negas

# join all images and save results to file
comm.Reduce(maps_T_loc,  maps_T_tot,  root=0)
comm.Reduce(maps_Z_loc,  maps_Z_tot,  root=0)
comm.Reduce(maps_P_loc,  maps_P_tot,  root=0)
comm.Reduce(maps_Vg_loc, maps_Vg_tot, root=0)
comm.Reduce(maps_Mg_loc, maps_Mg_tot, root=0)
comm.Reduce(maps_Mc_loc, maps_Mc_tot, root=0)
comm.Reduce(maps_Ms_loc, maps_Ms_tot, root=0)
comm.Reduce(maps_HI_loc, maps_HI_tot, root=0)
comm.Reduce(maps_ne_loc, maps_ne_tot, root=0)

# save files to disk
if myrank==0:
    print('')
    print('%.3e <  T [K] < %.3e'%(np.min(maps_T_tot),  np.max(maps_T_tot)))
    print('%.3e <   Z    < %.3e'%(np.min(maps_Z_tot),  np.max(maps_Z_tot)))
    print('%.3e <  Mgas  < %.3e'%(np.min(maps_Mg_tot), np.max(maps_Mg_tot)))
    print('%.3e <  Mcdm  < %.3e'%(np.min(maps_Mc_tot), np.max(maps_Mc_tot)))
    print('%.3e <  Mstar < %.3e'%(np.min(maps_Ms_tot), np.max(maps_Ms_tot)))
    print('%.3e <  Vgas  < %.3e'%(np.min(maps_Vg_tot), np.max(maps_Vg_tot)))
    print('%.3e <  M_HI  < %.3e'%(np.min(maps_HI_tot), np.max(maps_HI_tot)))
    print('%.3e <   P    < %.3e'%(np.min(maps_P_tot),  np.max(maps_P_tot)))
    print('%.3e <   n_e  < %.3e'%(np.min(maps_ne_tot), np.max(maps_ne_tot)))

    fout1 = '%s/Images_T_%s_z=%.2f.npy'%(root_out,sim,redshift)      
    fout2 = '%s/Images_Z_%s_z=%.2f.npy'%(root_out,sim,redshift)    
    fout3 = '%s/Images_Mgas_%s_z=%.2f.npy'%(root_out,sim,redshift)   
    fout4 = '%s/Images_Mcdm_%s_z=%.2f.npy'%(root_out,sim,redshift)   
    fout5 = '%s/Images_Mstar_%s_z=%.2f.npy'%(root_out,sim,redshift)  
    fout6 = '%s/Images_Vgas_%s_z=%.2f.npy'%(root_out,sim,redshift)   
    fout7 = '%s/Images_HI_%s_z=%.2f.npy'%(root_out,sim,redshift)     
    fout8 = '%s/Images_P_%s_z=%.2f.npy'%(root_out,sim,redshift)      
    fout9 = '%s/Images_ne_%s_z=%.2f.npy'%(root_out,sim,redshift)    
        
    np.save(fout1, maps_T_tot)
    np.save(fout2, maps_Z_tot)
    np.save(fout3, maps_Mg_tot)
    np.save(fout4, maps_Mc_tot)
    np.save(fout5, maps_Ms_tot)
    np.save(fout6, maps_Vg_tot)
    np.save(fout7, maps_HI_tot)
    np.save(fout8, maps_P_tot)
    np.save(fout9, maps_ne_tot)




import numpy as np
import sys,os

####################################### INPUT ###########################################
nodes_per_sim = 1
start         = 800   #first realization to do
end           = 1000   #last  realization do do
#########################################################################################

# do a loop over the different jobs
for i in xrange(start,end,1):
    
    script="""#!/bin/bash
#SBATCH --job-name=%d
#SBATCH -o OUTPUT.o%d
#SBATCH -e OUTPUT.e%d  
#SBATCH --partition=general
#####SBATCH --partition=preempt
#SBATCH --nodes=%d
#SBATCH --ntasks-per-node=48
#SBATCH --export=ALL
#SBATCH -t 7-00:00    

module purge
module load slurm
module load gcc
module load openmpi
module load lib/fftw2/2.1.5-openmpi1
module load lib/gsl
module load lib/hdf5

LPT_code="/simons/scratch/fvillaescusa/CAMEL/Software/2lpt/2LPTic"
G3="/simons/scratch/fvillaescusa/CAMEL/Codes/Gadget_III/P-Gadget3"
root="/simons/scratch/fvillaescusa/CAMEL/Sims/IllustrisTNG_DM/"

folder=$root"%d"
cd $folder

if [ ! -f snap_033.hdf5 ]; then

    if [ ! -f sim_stopped ]; then

        if [ ! -f restartfiles/restart.0 ]; then

           cd ICs
           mpirun -n 8 $LPT_code 2LPT.param >> logIC
           cd ..
           mpirun -n %d $G3 G3.param >> logfile    

        else

           mpirun -n %d $G3 G3.param 1 >> logfile    

        fi
    fi
fi
    """%(i, i, i, nodes_per_sim, i, nodes_per_sim*48, nodes_per_sim*48)

    # write the submission script inside each folder
    f = open('/simons/scratch/fvillaescusa/CAMEL/Sims/IllustrisTNG_DM/%d/script.sh'%i, 'w')
    f.write(script);  f.close()
    os.chdir('/simons/scratch/fvillaescusa/CAMEL/Sims/IllustrisTNG_DM/%d/'%i)
    os.system('sbatch script.sh')



# There were some problems with some simulations. We use this script to delete the data
# products of them, so we can rerun the pipeline cleanly
import numpy as np
import sys,os

#################################### INPUT ###########################################
root = '/mnt/ceph/users/camels/Results'
sim  = 'IllustrisTNG' 
wrong_realizations = ['LH_118', 'LH_142', 'LH_147', 'LH_185', 'LH_260', 'LH_261',
                      'LH_263', 'LH_291', 'LH_547', 'LH_563', 'LH_579', 'LH_596']
folders = ['baryon_fraction', 'baryon_fraction_SO', 'BH', 'HMF', 'Pk', 'Pk_ratio', 
           'Radii', 'SFR', 'SFRH', 'SMF', 'SO', 'Vmax']
######################################################################################

# do a loop over all the realizations that went wrong
for wr in wrong_realizations:

    # do a loop over the different folders
    for folder in folders:

        # get the name of the folder with the results
        fin = '%s/%s/%s/%s'%(root,folder,sim,wr)
        os.system('rm -rf %s'%fin)



from setuptools import setup
from setuptools import find_packages
from setuptools import Extension
import numpy

with open("README.md", "r") as f:
    documentation = f.read()

setup(
    name         = 'CAMELS_library',
    version      = "0.1", 
    author       = 'Francisco Villaescusa-Navarro',
    author_email = 'villaescusa.francisco@gmail.com',
    description  = "Python routines to analyze the data from the simulations of the Cosmology and Astrophysics with MachinE Learning Simulations (CAMELS) project",
    url          = "https://github.com/franciscovillaescusa/CAMELS",
    license      = 'MIT',
    long_description_content_type = "text/markdown",
    long_description = documentation,
    packages=find_packages(where="library/"),
    include_dirs=[numpy.get_include()],
    install_requires=['scipy', 'h5py', 'Pylians'],
    package_dir={'':'library/'},
    py_modules=['camels_library']
)







import numpy as np
import torch
import sys,os
sys.path.append('../')
import data as data
import architecture

#################################### INPUT ##########################################
# data parameters
root_in      = '/mnt/ceph/users/camels'
root_out     = '/mnt/ceph/users/camels/Results/neural_nets/SFRH_2_params'
sim          = 'IllustrisTNG'
seed         = 1
realizations = 1000
bins_SFRH    = 100

# architecture parameters
h1 = 750
h2 = 750
h3 = 750
dropout_rate = 0.0

# training parameters
batch_size = 10

minimum = np.array([0.1, 0.6, 0.25, 0.25, 0.5, 0.5])
width   = np.array([0.4, 0.4, 3.75, 3.75, 1.5, 1.5])

# name of output files
name   = '3hd_750_750_750_0.0_3e-3'
fout   = '%s/results/%s.txt'%(root_out,name)
fmodel = '%s/models/%s.pt'%(root_out,name)
#####################################################################################

# get GPU if available
if torch.cuda.is_available():
    print("CUDA Available")
    device = torch.device('cuda')
else:
    print("CUDA Not Available")
    device = torch.device('cpu')

# get the test dataset
test_loader  = data.create_dataset('test', seed, realizations, root_in, bins_SFRH, 
                                   sim, batch_size, root_out)

# get the number of elements in the test set
size = 0
for params_test, SFRH_test in test_loader:
    size += params_test.shape[0]

# define the array with the results
pred = np.zeros((size,6), dtype=np.float32)
true = np.zeros((size,6), dtype=np.float32)

# get the parameters of the trained model
#model = architecture.model_1hl(bins_SFRH, h1, 6, dropout_rate)
#model = architecture.model_2hl(bins_SFRH, h1, h2, 6, dropout_rate)
model = architecture.model_3hl(bins_SFRH, h1, h2, h3, 6, dropout_rate)
model.load_state_dict(torch.load(fmodel))
model.to(device=device)

# loop over the different batches and get the prediction
offset = 0
model.eval()
for params_test, SFRH_test in test_loader:
    with torch.no_grad():
        #params_test = params_test.to(device)
        SFRH_test   = SFRH_test.to(device)
        params_pred = model(SFRH_test)
        length = params_test.shape[0]
        pred[offset:offset+length] = params_pred.cpu().numpy()
        true[offset:offset+length] = params_test.numpy()
        offset += length

# compute the rmse; de-normalize
error_norm = ((pred - true))**2
pred  = pred*width + minimum
true  = true*width + minimum
error = (pred - true)**2

print('Error^2 norm      = %.3e'%np.mean(error_norm))
print('Error             = %.3e'%np.sqrt(np.mean(error)))
print('Relative error Om = %.3e'%np.sqrt(np.mean(error[:,0])))
print('Relative error s8 = %.3e'%np.sqrt(np.mean(error[:,1])))
print('Relative error A1 = %.3e'%np.sqrt(np.mean(error[:,2])))
print('Relative error A2 = %.3e'%np.sqrt(np.mean(error[:,3])))
print('Relative error A3 = %.3e'%np.sqrt(np.mean(error[:,4])))
print('Relative error A4 = %.3e'%np.sqrt(np.mean(error[:,5])))

# save results to file
results = np.zeros((size,12))
results[:,0:6]  = true
results[:,6:12] = pred
np.savetxt(fout, results)




from mpi4py import MPI
import numpy as np
import sys,os
import camb
from pyDOE import *

###### MPI DEFINITIONS ######                                    
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

#################################### INPUT #############################################
root_out   = '/simons/scratch/fvillaescusa/CAMEL/Sims/IllustrisTNG_DM'
dimensions = 6     #number of cosmo+astro parameters
points     = 1000  #number of simulations
seed       = 1     #random seed for latin-hypercube
fout_lh    = 'latin_hypercube_params.txt'

# CAMB parameters
Omega_b      = 0.049
h            = 0.6711
ns           = 0.9624
hierarchy    = 'degenerate'
Mnu          = 0.0 #eV
Nnu          = 0   #number of massive neutrinos
Neff         = 3.046
As           = 2.13e-9
tau          = None
Omega_k      = 0.0
pivot_scalar = 0.05
pivot_tensor = 0.05
kmax         = 200.0
k_per_logint = 20
redshifts    = [0]

#              [Om,   s8,   astro1, astro2, astro3, astro4]
Min = np.array([0.1,  0.6,  0.25,   0.25,   0.5,    0.5])
Max = np.array([0.5,  1.0,  4.00,   4.00,   2.0,    2.0])
########################################################################################

os.system('source ~/gadget_env_popeye')

# set logscale for astro params
Min[2:] = np.log10(Min[2:])
Max[2:] = np.log10(Max[2:])

# get latin hypercube
np.random.seed(seed)
coords = lhs(dimensions, samples=points, criterion='c')

# generate the parameters
params = np.zeros((points,dimensions), dtype=np.float64)
params = Min + coords*(Max-Min)
params[:,2:] = 10**params[:,2:]
print '%.5f < Om < %.5f : <Om> = %.5f'\
    %(np.min(params[:,0]), np.max(params[:,0]), np.mean(params[:,0]))
print '%.5f < s8 < %.5f : <s8> = %.5f'\
    %(np.min(params[:,1]), np.max(params[:,1]), np.mean(params[:,1]))
print '%.5f < Astro1 < %.5f : <Astro1> = %.5f'\
    %(np.min(params[:,2]), np.max(params[:,2]), 10**(np.mean(np.log10(params[:,2]))))
print '%.5f < Astro2 < %.5f : <Astro2> = %.5f'\
    %(np.min(params[:,3]), np.max(params[:,3]), 10**(np.mean(np.log10(params[:,3]))))
print '%.5f < Astro3 < %.5f : <Astro3> = %.5f'\
    %(np.min(params[:,4]), np.max(params[:,4]), 10**(np.mean(np.log10(params[:,4]))))
print '%.5f < Astro4 < %.5f : <Astro4> = %.5f'\
    %(np.min(params[:,5]), np.max(params[:,5]), 10**(np.mean(np.log10(params[:,5]))))

# check if the generated parameters are the same as the saved ones
if os.path.exists(fout_lh):  
    params_file = np.loadtxt(fout_lh)
    equal = (np.around(params,5)==params_file) #approximate params with 5 decimals
    if np.any(equal==False):  raise Exception('Cosmological parameters are different!!!')
else:
    np.savetxt(fout_lh, params, fmt='%.5f')


# get the numbers each cpu will work on
numbers = np.where(np.arange(points)%nprocs==myrank)[0]

# do a loop over all the points in the latin hypercube
for i in numbers:

    # create output folder in case it does not exists
    folder = '%s/%s/'%(root_out,i)
    if not(os.path.exists(folder)):  os.system('mkdir %s'%folder)

    # create initial conditions folder if it does not exists
    folder_ICs = '%s/ICs'%folder 
    if not(os.path.exists(folder_ICs)):  os.system('mkdir %s'%folder_ICs)

    # find the values of the cosmological parameters
    Omega_m = params[i,0]
    s8      = params[i,1]
    astro1  = params[i,2]
    astro2  = params[i,3]
    astro3  = params[i,4]
    astro4  = params[i,5]

    g = open('%s/CosmoAstro_params.txt'%folder, 'w')
    g.write('%.5f %.5f %.5f %.5f %.5f %.5f\n'\
            %(Omega_m, s8, astro1, astro2, astro3, astro4))
    g.close()

    print 'realization %d'%i
    print 'Omega_m = %.5f'%Omega_m
    print 's8      = %.5f'%s8
    print 'Astro1  = %.5f'%astro1
    print 'Astro2  = %.5f'%astro2
    print 'Astro3  = %.5f'%astro3
    print 'Astro4  = %.5f'%astro4

    ##### run CAMB #####
    Omega_c  = Omega_m - Omega_b
    pars     = camb.CAMBparams()

    # set accuracy of the calculation
    pars.set_accuracy(AccuracyBoost=4.0, lSampleBoost=4.0, lAccuracyBoost=4.0, 
                      HighAccuracyDefault=True, DoLateRadTruncation=True)

    # set value of the cosmological parameters
    pars.set_cosmology(H0=h*100.0, ombh2=Omega_b*h**2, omch2=Omega_c*h**2, 
                       mnu=Mnu, omk=Omega_k, neutrino_hierarchy=hierarchy, 
                       num_massive_neutrinos=Nnu, nnu=Neff, tau=tau)
                   
    # set the value of the primordial power spectrum parameters
    pars.InitPower.set_params(As=As, ns=ns, 
                              pivot_scalar=pivot_scalar, pivot_tensor=pivot_tensor)

    # set redshifts, k-range and k-sampling
    pars.set_matter_power(redshifts=redshifts, kmax=kmax, k_per_logint=k_per_logint)

    # compute results
    results = camb.get_results(pars)

    # save parameter values to file
    f = open('%s/CAMB.params'%folder_ICs,'w');  f.write('%s'%pars);  f.close()

    # interpolate to get Pmm, Pcc...etc
    k, zs, Pkmm = results.get_matter_power_spectrum(minkh=2e-5, maxkh=kmax, 
                                                    npoints=400, var1=7, var2=7, 
                                                    have_power_spectra=True, 
                                                    params=None)

    # do a loop over all redshifts
    for j,z in enumerate(zs):
        fout = '%s/Pk_m_z=%.3f.txt'%(folder_ICs,z)
        np.savetxt(fout, np.transpose([k,Pkmm[j,:]]))




    ########################## write 2LPT parameter file #######################
    # parameter file for standard simulations
    a="""
Nmesh            512
Nsample          256       
Box              25000.0   
FileBase         ics         
OutputDir        ./
GlassFile        /simons/scratch/fvillaescusa/CAMEL/Software/2lpt/GLASS/dummy_glass_dmonly_64.dat   
GlassTileFac     4         
Omega            %.4f    
OmegaLambda      %.4f    
OmegaBaryon      0.00
OmegaDM_2ndSpecies  0.0    
HubbleParam      0.6711    
Redshift         127       
Sigma8           %.4f       
SphereMode       0         
WhichSpectrum    2         
FileWithInputSpectrum   ./Pk_m_z=0.000.txt
InputSpectrum_UnitLength_in_cm  3.085678e24 
ShapeGamma       0.201     
PrimordialIndex  1.0       
    
Phase_flip          0      
RayleighSampling    1      
Seed                %d      
    
NumFilesWrittenInParallel 8  
UnitLength_in_cm          3.085678e21  
UnitMass_in_g             1.989e43     
UnitVelocity_in_cm_per_s  1e5          
    
WDM_On               0      
WDM_Vtherm_On        0      
WDM_PartMass_in_kev  10.0   
    """%(Omega_m, 1.0-Omega_m, s8, i)

    # save parameters to file                                    
    f = open('%s/2LPT.param'%folder_ICs, 'w');  f.write(a);  f.close()



    ########################## write G3 parameter file #######################
    G3_params="""
InitCondFile              ./ICs/ics
OutputDir                 ./
OutputListFilename        /simons/scratch/fvillaescusa/CAMEL/Sims/times.txt
NumFilesPerSnapshot       1
NumFilesWrittenInParallel 1
CpuTimeBetRestartFile     10800.0   
TimeLimitCPU              10000000  
ICFormat                  1
SnapFormat                3
TimeBegin                 0.0078125     
TimeMax	                  1.00          
Omega0	                  %.5f    
OmegaLambda               %.5f
OmegaBaryon               0.0000     
HubbleParam               0.6711     
BoxSize                   25000.0

SofteningGas              0.0
SofteningHalo             0.5   
SofteningDisk             0.0
SofteningBulge            0.0
SofteningStars            0.0
SofteningBndry            0.0
SofteningGasMaxPhys       0.0
SofteningHaloMaxPhys      0.5
SofteningDiskMaxPhys      0.0
SofteningBulgeMaxPhys     0.0
SofteningStarsMaxPhys     0.0
SofteningBndryMaxPhys     0.0

PartAllocFactor           2.5  
MaxMemSize	          15500
BufferSize                300
CoolingOn                 0
StarformationOn           0

TypeOfTimestepCriterion   0   	                    
ErrTolIntAccuracy         0.025  
MaxSizeTimestep           0.005
MinSizeTimestep           0.0

ErrTolTheta               0.5
TypeOfOpeningCriterion    1
ErrTolForceAcc            0.005
TreeDomainUpdateFrequency 0.01
 
DesNumNgb                 33
MaxNumNgbDeviation        2
ArtBulkViscConst          1.0
InitGasTemp               273.0  
MinGasTemp                10.0    
CourantFac                0.15
ComovingIntegrationOn     1    
PeriodicBoundariesOn      1    
MinGasHsmlFractional      0.1  
OutputListOn              1    
TimeBetSnapshot           1.   
TimeOfFirstSnapshot       1.   
TimeBetStatistics         0.5  
MaxRMSDisplacementFac     0.25 
EnergyFile                energy.txt
InfoFile                  info.txt
TimingsFile               timings.txt
CpuFile                   cpu.txt
TimebinFile               Timebin.txt
SnapshotFileBase          snap
RestartFile               restart
ResubmitOn                0
ResubmitCommand           /home/vspringe/autosubmit
UnitLength_in_cm          3.085678e21      
UnitMass_in_g             1.989e43         
UnitVelocity_in_cm_per_s  1e5              
GravityConstantInternal   0
    """%(Omega_m, 1.0-Omega_m)


    # save G3 parameters to file                                    
    f = open('%s/G3.param'%folder, 'w');  f.write(G3_params);  f.close()



import numpy as np
import sys,os

####################################### INPUT ###########################################
nodes_per_sim = 1
start         = 0   #350   #first realization to do
end           = 87 #400   #last  realization do do
#########################################################################################
# 0-10
#11-21
#22-32
#33-43
#44-54
#55-65

#numbers = np.array([1501, 1503, 1505, 1507, 1509,
#                    1512, 1514,       1518, 1520,
#                    1523, 1525,       1529, 1531,
#                    1534, 1536,       1540, 1542,
#                    1545, 1547,       1551, 1553,
#                    1556, 1558,       1562, 1564])
numbers = np.array(['1505_0', '1505_1', '1505_2', '1505_3', '1505_4'])

# do a loop over the different jobs
for i in numbers:
    
    script="""#!/bin/bash
#SBATCH --job-name=%s
#SBATCH -o OUTPUT.o%s
#SBATCH -e OUTPUT.e%s  
#SBATCH --partition=general
#####SBATCH --partition=preempt
#SBATCH --nodes=%d
#SBATCH --ntasks-per-node=48
#SBATCH --export=ALL
#SBATCH -t 7-00:00    
#####SBATCH --mail-type=ALL
#####SBATCH --mail-user=villaescusa.francisco@gmail.com

module purge
module load slurm
module load gcc
module load openmpi
module load lib/fftw2/2.1.5-openmpi1
module load lib/gsl
module load lib/hdf5
export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/home/dangles/local/grackle_popeye/lib
LPT_code="/simons/scratch/fvillaescusa/hydro_DL/Software/2lpt/2LPTic"

root="/simons/scratch/fvillaescusa/hydro_DL/SIMBA/"

folder=$root"%s"
cd $folder

if [ ! -f snap_033.hdf5 ]; then

    #if [ ! -f restartfiles/restart.0 ]; then

        cd ICs
        mpirun -n 8 $LPT_code 2LPT.param >> logIC
        cd ..
        mpirun -n %d ./gizmo-mufasa/GIZMO GIZMO.param >> logfile    

        #else

        #mpirun -n %d ./gizmo-mufasa/GIZMO GIZMO.param 1 >> logfile    

    #fi
fi
    """%(i, i, i, nodes_per_sim, i, nodes_per_sim*48, nodes_per_sim*48)

    # write the submission script inside each folder
    f = open('/simons/scratch/fvillaescusa/hydro_DL/SIMBA/%s/script.sh'%i, 'w')
    f.write(script);  f.close()
    os.chdir('/simons/scratch/fvillaescusa/hydro_DL/SIMBA/%s/'%i)
    os.system('sbatch script.sh')



# This script creates 2D images/fields from different fields of the simulations
from mpi4py import MPI
import numpy as np
import sys,os,h5py
import MAS_library as MASL
sys.path.append('../analysis')
import camel_library as CL

###### MPI DEFINITIONS ###### 
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

#################################### INPUT ##########################################
# data parameters
root_in       = '/mnt/ceph/users/camels'
root_out      = '/mnt/ceph/users/camels/Results/images_LH'
TREECOOL_file = '/mnt/ceph/users/camels/Software/TREECOOL'
sim           = 'IllustrisTNG'
snapnum       = 31
realizations  = 1000

# images parameters
grid         = 250
splits       = 5

# parameters of the density field routine
periodic     = True
verbose      = False
x_min, y_min = 0.0, 0.0
#####################################################################################

for snapnum in np.arange(0,34):

    # find the numbers that each cpu will work with
    numbers = np.where(np.arange(realizations)%nprocs==myrank)[0]

    # define the matrix hosting all the maps
    maps_T_loc  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_T_tot  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Z_loc  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Z_tot  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_P_loc  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_P_tot  = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Mg_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Mg_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Mc_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Mc_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Ms_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Ms_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Vg_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_Vg_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_HI_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_HI_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_ne_loc = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)
    maps_ne_tot = np.zeros((splits*3*realizations, grid, grid), dtype=np.float32)

    # do a loop over all realizations
    for i in numbers:

        if myrank==0:  print(i, end=' ', flush=True)

        # read the gas positions, radii and temperature of gas particles
        snapshot = '%s/Sims/%s/%d/snap_%03d.hdf5'%(root_in,sim,i,snapnum)
        if not(os.path.exists(snapshot)):  raise Exception('%s not found'%snapshot)
        f = h5py.File(snapshot, 'r')
        BoxSize = f['Header'].attrs[u'BoxSize']/1e3 #Mpc/h
        redshift = f['Header'].attrs[u'Redshift'];  
        Masses  = f['Header'].attrs[u'MassTable'][:]*1e10
        pos_g   = f['PartType0/Coordinates'][:]/1e3;  pos_g = pos_g.astype(np.float32)#Mpc/h
        pos_c   = f['PartType1/Coordinates'][:]/1e3;  pos_c = pos_c.astype(np.float32)#Mpc/h
        pos_s   = f['PartType4/Coordinates'][:]/1e3;  pos_s = pos_s.astype(np.float32)#Mpc/h
        vel_g   = f['/PartType0/Velocities'][:]/np.sqrt(1.0+redshift)
        vel_g   = np.sqrt(vel_g[:,0]**2 + vel_g[:,1]**2 + vel_g[:,2]**2)
        Mg      = f['PartType0/Masses'][:]*1e10     #Msun/h
        Ms      = f['PartType4/Masses'][:]*1e10     #Msun/h
        if sim=='IllustrisTNG':
            Mc  = np.ones(pos_c.shape[0], dtype=np.float32)*Masses[1] #Msun/h
            Rg  = f['PartType0/SubfindHsml'][:]/1e3  #Mpc/h
            Rc  = f['PartType1/SubfindHsml'][:]/1e3  #Mpc/h
            Rs  = f['PartType4/SubfindHsml'][:]/1e3  #Mpc/h
            Z   = f['/PartType0/GFM_Metallicity'][:] #metallicity
            HI  = CL.HI_mass(snapshot, TREECOOL_file)
        else:
            Mc  = f['PartType1/Masses'][:]*1e10       #Msun/h
            Rg  = f['PartType0/AGS-Softening'][:]/1e3 #Mpc/h
            Rc  = f['PartType1/AGS-Softening'][:]/1e3 #Mpc/h
            Rs  = f['PartType4/AGS-Softening'][:]/1e3 #Mpc/h
            HI  = f['/PartType0/NeutralHydrogenAbundance'][:]  #HI/H
        f.close()
        T  = CL.temperature(snapshot)      #K
        P  = CL.pressure(snapshot)         #Msun*(km/s)^2/kpc^3
        ne = CL.electron_density(snapshot) #1e20 electrons*h^2/Mpc^3

        # do a loop over the three different axes
        for axis in [0,1,2]:

            axis_x, axis_y = (axis+1)%3, (axis+2)%3 

            # do a loop over the different slices of each axis
            for j in range(splits):

                # get the number of the map; this is important for MPI
                num = i*3*splits + axis*splits + j

                # find the range in the slice
                minimum, maximum = j*BoxSize/splits, (j+1)*BoxSize/splits

                # select the particles in the considered slice
                indexes_g = np.where((pos_g[:,axis]>=minimum) & (pos_g[:,axis]<maximum))[0] 
                pos_g_    = pos_g[indexes_g]
                vel_g_    = vel_g[indexes_g]
                Rg_       = Rg[indexes_g]
                Mg_       = Mg[indexes_g]
                T_        = T[indexes_g]
                Z_        = Z[indexes_g]
                HI_       = HI[indexes_g]
                P_        = P[indexes_g]
                ne_       = ne[indexes_g]
                indexes_c = np.where((pos_c[:,axis]>=minimum) & (pos_c[:,axis]<maximum))[0] 
                pos_c_    = pos_c[indexes_c]
                Mc_       = Mc[indexes_c]
                Rc_       = Rc[indexes_c]
                indexes_s = np.where((pos_s[:,axis]>=minimum) & (pos_s[:,axis]<maximum))[0] 
                pos_s_    = pos_s[indexes_s]
                Ms_       = Ms[indexes_s]
                Rs_       = Rs[indexes_s]

                # project gas mass*temperatures into a 2D map
                TM = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(TM, pos_g_, T_*Mg_, Rg_, x_min, y_min, 
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                # project gas mass*metallicity into a 2D map
                ZM = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(ZM, pos_g_, Z_*Mg_, Rg_, x_min, y_min, 
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                # project gas velocity*mass into a 2D map
                VM = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(VM, pos_g_, vel_g_*Mg_, Rg_, x_min, y_min,
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                # project gas pressure*mass into a 2D map
                PM = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(PM, pos_g_, P_*Mg_, Rg_, x_min, y_min,
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                # project gas mass into a 2D map
                Mgas = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(Mgas, pos_g_, Mg_, Rg_, x_min, y_min,
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                # project cdm mass into a 2D map
                Mcdm = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(Mcdm, pos_c_, Mc_, Rc_, x_min, y_min,
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                # project stellar mass into a 2D map
                Mstar = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(Mstar, pos_s_, Ms_, Rs_, x_min, y_min,
                                   axis_x, axis_y, BoxSize, periodic, verbose)
                
                # project HI mass into a 2D map
                HIgas = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(HIgas, pos_g_, HI_, Rg_, x_min, y_min,
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                # project electron density into a 2D map
                negas = np.zeros((grid,grid), dtype=np.float64)
                MASL.voronoi_RT_2D(negas, pos_g_, ne_, Rg_, x_min, y_min,
                                   axis_x, axis_y, BoxSize, periodic, verbose)

                maps_T_loc[num]  = TM/Mgas
                maps_Z_loc[num]  = ZM/Mgas
                maps_P_loc[num]  = PM/Mgas
                maps_Vg_loc[num] = VM/Mgas
                maps_Mg_loc[num] = Mgas
                maps_Mc_loc[num] = Mcdm
                maps_Ms_loc[num] = Mstar
                maps_HI_loc[num] = HIgas
                maps_ne_loc[num] = negas

    # join all images and save results to file
    comm.Reduce(maps_T_loc,  maps_T_tot,  root=0)
    comm.Reduce(maps_Z_loc,  maps_Z_tot,  root=0)
    comm.Reduce(maps_P_loc,  maps_P_tot,  root=0)
    comm.Reduce(maps_Vg_loc, maps_Vg_tot, root=0)
    comm.Reduce(maps_Mg_loc, maps_Mg_tot, root=0)
    comm.Reduce(maps_Mc_loc, maps_Mc_tot, root=0)
    comm.Reduce(maps_Ms_loc, maps_Ms_tot, root=0)
    comm.Reduce(maps_HI_loc, maps_HI_tot, root=0)
    comm.Reduce(maps_ne_loc, maps_ne_tot, root=0)

    # save files to disk
    if myrank==0:
        print('')
        print('%.3e <  T [K] < %.3e'%(np.min(maps_T_tot),  np.max(maps_T_tot)))
        print('%.3e <   Z    < %.3e'%(np.min(maps_Z_tot),  np.max(maps_Z_tot)))
        print('%.3e <  Mgas  < %.3e'%(np.min(maps_Mg_tot), np.max(maps_Mg_tot)))
        print('%.3e <  Mcdm  < %.3e'%(np.min(maps_Mc_tot), np.max(maps_Mc_tot)))
        print('%.3e <  Mstar < %.3e'%(np.min(maps_Ms_tot), np.max(maps_Ms_tot)))
        print('%.3e <  Vgas  < %.3e'%(np.min(maps_Vg_tot), np.max(maps_Vg_tot)))
        print('%.3e <  M_HI  < %.3e'%(np.min(maps_HI_tot), np.max(maps_HI_tot)))
        print('%.3e <   P    < %.3e'%(np.min(maps_P_tot),  np.max(maps_P_tot)))
        print('%.3e <   n_e  < %.3e'%(np.min(maps_ne_tot), np.max(maps_ne_tot)))

        fout1 = '%s/Images_T_%s_z=%.2f.npy'%(root_out,sim,redshift)      
        fout2 = '%s/Images_Z_%s_z=%.2f.npy'%(root_out,sim,redshift)    
        fout3 = '%s/Images_Mgas_%s_z=%.2f.npy'%(root_out,sim,redshift)   
        fout4 = '%s/Images_Mcdm_%s_z=%.2f.npy'%(root_out,sim,redshift)   
        fout5 = '%s/Images_Mstar_%s_z=%.2f.npy'%(root_out,sim,redshift)  
        fout6 = '%s/Images_Vgas_%s_z=%.2f.npy'%(root_out,sim,redshift)   
        fout7 = '%s/Images_HI_%s_z=%.2f.npy'%(root_out,sim,redshift)     
        fout8 = '%s/Images_P_%s_z=%.2f.npy'%(root_out,sim,redshift)      
        fout9 = '%s/Images_ne_%s_z=%.2f.npy'%(root_out,sim,redshift)    
        
        np.save(fout1, maps_T_tot)
        np.save(fout2, maps_Z_tot)
        np.save(fout3, maps_Mg_tot)
        np.save(fout4, maps_Mc_tot)
        np.save(fout5, maps_Ms_tot)
        np.save(fout6, maps_Vg_tot)
        np.save(fout7, maps_HI_tot)
        np.save(fout8, maps_P_tot)
        np.save(fout9, maps_ne_tot)

    comm.Barrier()



# This script creates temperature images from the 27 IllustrisTNG fiducial realizations
from mpi4py import MPI
import numpy as np
import sys,os,h5py
import MAS_library as MASL
sys.path.append('../analysis')
import camel_library as CL

###### MPI DEFINITIONS ###### 
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

#################################### INPUT ##########################################
root_in      = '/mnt/ceph/users/camels'
root_out     = '/mnt/ceph/users/camels/Results/autoencoder'
sim          = 'IllustrisTNG' #'IllustrisTNG'
snapnum      = 33
realizations = 27

grid         = 250
splits       = 5

# parameters of the density field routine
periodic = True
x_min, y_min = 0.0, 0.0
#####################################################################################

# compute the number of images that fit in each direction
# each subimage will have 64x64 pixels out of grid x grid pixels
# with no overalping images we can have (grid/64)^2 subimages
# however, we can have much more subimages if we along overlaping
# this parameter determines how much the frame moves when changing the image
# for non-overlapping images this will be equal to 64
subimages_dir = (grid-64)//offset_imag + 1

# compute the total number of images to produce
images_tot = subimages_dir*subimages_dir*splits*3

# define the matrix hosting all the maps
maps_local = np.zeros((realizations*splits*3, grid, grid), dtype=np.float32)
maps_total = np.zeros((realizations*splits*3, grid, grid), dtype=np.float32)

# find the numbers that each cpu will work with
numbers = np.where(np.arange(realizations)%nprocs==myrank)[0]

# do a loop over all realizations
for i in numbers:

    # read the gas positions, radii and temperature of gas particles
    snapshot = '%s/Sims/%s/1505_%d/snap_%03d.hdf5'%(root_in,sim,i,snapnum)
    f = h5py.File(snapshot, 'r')
    BoxSize = f['Header'].attrs[u'BoxSize']/1e3 #Mpc/h
    pos     = f['PartType0/Coordinates'][:]/1e3 #Mpc/h
    mass    = f['PartType0/Masses'][:]*1e10     #Msun/h
    if sim=='IllustrisTNG':
        radius  = f['PartType0/SubfindHsml'][:]/1e3 #Mpc/h
    else:
        #radius  = f['PartType0/SmoothingLength'][:]/1e3 #Mpc/h
        radius  = f['PartType0/AGS-Softening'][:]/1e3 #Mpc/h
    f.close()
    T = CL.temperature(snapshot)

    # do a loop over the three different axes
    for axis in [0,1,2]:

        axis_x, axis_y = (axis+1)%3, (axis+2)%3 

        # do a loop over the different slices of each axis
        for j in range(splits):

            # get the index of the map
            count = splits*3*i + splits*axis + j

            # find the range in the slice
            minimum, maximum = j*BoxSize/splits, (j+1)*BoxSize/splits

            # select the particles in the considered slice
            indexes = np.where((pos[:,axis]>=minimum) & (pos[:,axis]<maximum))[0] 
            pos_slice    = pos[indexes].astype(np.float32)
            radius_slice = radius[indexes]
            mass_slice   = mass[indexes]
            T_slice      = T[indexes]

            # project particle mass*temperatures into a 2D maps
            TM = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(TM, pos_slice, T_slice*mass_slice, radius_slice, 
                    x_min, y_min, axis_x, axis_y, BoxSize, periodic, verbose=False)

            # project particle mass into a 2D maps
            M = np.zeros((grid,grid), dtype=np.float64)
            MASL.voronoi_RT_2D(M, pos_slice, mass_slice, radius_slice, x_min, y_min,
                               axis_x, axis_y, BoxSize, periodic, verbose=False)

            maps_local[count] = TM/M
    print('Realization %d : %d maps'%(i, count+1))

# join all images and save results to file
comm.Reduce(maps_local, maps_total, root=0)

if myrank==0:
    print(maps_total.shape)
    print('%.3e < T [K] < %.3e'%(np.min(maps_total), np.max(maps_total)))
    np.save('%s/Images_T_fiducial_IllustrisTNG.npy'%root_out, maps_total)
